# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

* :class:`ChartTemplatesToolBar`
* :class:`FindToolbar`
* :class:`FormatToolbar`
* :class:`MacroToolbar`
* :func:`add_toolbutton_widget`

"""

from typing import Tuple

from PyQt5.QtWidgets import (
        QToolBar, QToolButton, QMenu, QWidget, QHBoxLayout, QUndoView,
        QMainWindow)

try:
    import matplotlib.figure as matplotlib_figure
except ImportError:
    matplotlib_figure = None

try:
    import rpy2
    from rpy2.robjects.packages import importr, PackageNotInstalledError
except ImportError:
    rpy2 = None

try:
    import plotnine
except ImportError:
    plotnine = None

try:
    from pyspread.actions import MainWindowActions, ChartDialogActions
    from pyspread.icons import Icon
    from pyspread.menus import ToolbarManagerMenu
    from pyspread.widgets import FindEditor
except ImportError:
    from actions import MainWindowActions, ChartDialogActions
    from icons import Icon
    from menus import ToolbarManagerMenu
    from widgets import FindEditor


def add_toolbutton_widget(button: QWidget, widget: QWidget,
                          minsize: Tuple[int, int] = (300, 200),
                          popup_mode: QToolButton.ToolButtonPopupMode
                          = QToolButton.MenuButtonPopup):
    """Adds a widget as menu to a tool_button

    :param button: Tool button for menu
    :param widget: Toolbar widget
    :param minsize: Minimum menu size
    :param popup_mode: Describes how the menu is popped up

    """

    button.setPopupMode(popup_mode)
    menu = QMenu(button)
    menu.setMinimumSize(*minsize)
    button.setMenu(menu)
    menu_layout = QHBoxLayout()
    menu_layout.addWidget(widget)
    menu.setLayout(menu_layout)
    menu.layout()


class ToolBarBase(QToolBar):
    """Base toolbar class that provides toolbar manager button method"""

    def add_widget(self, widget: QWidget):
        """Adds widget with addWidget and assigns action text and icon

        The widget must have a label attribute and an icon method.

        :param widget: Widget to be added

        """

        self.addWidget(widget)
        self.actions()[-1].setText(widget.label)
        self.actions()[-1].setIcon(widget.icon())

    def get_manager_button(self) -> QToolButton:
        """Returns QToolButton for managing the toolbar"""

        button = QToolButton(self)
        button.setText("Add/remove toolbar icons")
        button.setMenu(ToolbarManagerMenu(self))
        button.setIcon(Icon.menu_manager)
        button.setFixedWidth(int(button.height()/3))
        button.setPopupMode(QToolButton.InstantPopup)

        return button


class MainToolBar(ToolBarBase):
    """The main toolbar"""

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        self.main_window = main_window
        super().__init__("Main toolbar", main_window)

        self.setObjectName("Main Toolbar")
        self._create_toolbar(main_window.main_window_toolbar_actions)

    def _create_toolbar(self, actions: MainWindowActions):
        """Fills the main toolbar with QActions

        :param actions: Main window actions

        """

        self.addAction(actions.new)
        self.addAction(actions.open)
        self.addAction(actions.save)
        self.addAction(actions.export)

        self.addSeparator()

        self.addAction(actions.print)

        self.addSeparator()

        self.addAction(actions.undo)
        undo_button = self.widgetForAction(actions.undo)
        undo_view = QUndoView(self.main_window.undo_stack)
        add_toolbutton_widget(undo_button, undo_view)

        self.addAction(actions.redo)

        self.addSeparator()

        self.addAction(actions.cut)
        self.addAction(actions.copy)
        self.addAction(actions.copy_results)
        self.addAction(actions.paste)
        self.addAction(actions.paste)

        self.addSeparator()

        self.addAction(actions.sort_ascending)
        self.addAction(actions.sort_descending)

        self.addSeparator()

        self.addAction(actions.toggle_spell_checker)

        self.addWidget(self.get_manager_button())


class MacroToolbar(ToolBarBase):
    """The macro toolbar for pyspread"""

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        super().__init__("Macro toolbar", main_window)

        self.setObjectName("Macro toolbar")
        self._create_toolbar(main_window.main_window_toolbar_actions)

    def _create_toolbar(self, actions: MainWindowActions):
        """Fills the macro toolbar with QActions

        :param actions: Main window actions

        """

        self.addAction(actions.insert_image)
        if matplotlib_figure is not None:
            self.addAction(actions.insert_chart)

        self.addSeparator()

        self.addAction(actions.insert_sum)

        self.addWidget(self.get_manager_button())


class FindToolbar(ToolBarBase):
    """The find toolbar for pyspread"""

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        super().__init__("Find Toolbar", main_window)

        self.main_window = main_window
        self.setObjectName("Find Toolbar")
        self._create_toolbar(main_window.main_window_toolbar_actions)

    def _create_toolbar(self, actions: MainWindowActions):
        """Fills the find toolbar with QActions

        :param actions: Main window actions

        """

        self.find_editor = FindEditor(self)

        self.add_widget(self.find_editor)

        self.addAction(actions.find)
        self.addAction(actions.replace)

        self.addWidget(self.get_manager_button())


class FormatToolbar(ToolBarBase):
    """The format toolbar for pyspread"""

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        super().__init__("Format Toolbar", main_window)

        self.main_window = main_window
        self.setObjectName("Format Toolbar")
        self._create_toolbar(main_window.main_window_toolbar_actions)

    def _create_toolbar(self, actions: MainWindowActions):
        """Fills the format toolbar with QActions

        :param actions: Main window actions

        """

        menubar = self.main_window.menuBar()

        self.add_widget(self.main_window.widgets.font_combo)
        self.add_widget(self.main_window.widgets.font_size_combo)

        self.addAction(actions.bold)
        self.addAction(actions.italics)
        self.addAction(actions.underline)
        self.addAction(actions.strikethrough)

        self.addSeparator()

        self.add_widget(self.main_window.widgets.renderer_button)

        self.addSeparator()

        self.addAction(actions.freeze_cell)
        self.addAction(actions.lock_cell)
        self.addAction(actions.button_cell)

        self.addSeparator()

        self.addAction(actions.merge_cells)

        self.addSeparator()

        self.add_widget(self.main_window.widgets.rotate_button)
        self.add_widget(self.main_window.widgets.justify_button)
        self.add_widget(self.main_window.widgets.align_button)

        self.addSeparator()

        self.border_menu_button = QToolButton(self)
        self.border_menu_button.setText("Borders")
        self.border_menu_button.label = "Borders"
        border_submenu = menubar.format_menu.border_submenu
        self.border_menu_button.setMenu(border_submenu)
        self.border_menu_button.setIcon(Icon.border_menu)
        self.add_widget(self.border_menu_button)
        self.border_menu_button.setPopupMode(QToolButton.InstantPopup)

        self.line_width_button = QToolButton(self)
        self.line_width_button.setText("Border Width")
        self.line_width_button.label = "Border Width"
        line_width_submenu = menubar.format_menu.line_width_submenu
        self.line_width_button.setMenu(line_width_submenu)
        self.line_width_button.setIcon(Icon.format_borders)
        self.add_widget(self.line_width_button)
        self.line_width_button.setPopupMode(QToolButton.InstantPopup)

        self.addSeparator()

        text_color_button = self.main_window.widgets.text_color_button
        text_color_button.set_max_size(self.iconSize())
        self.add_widget(text_color_button)

        line_color_button = self.main_window.widgets.line_color_button
        line_color_button.set_max_size(self.iconSize())
        self.add_widget(line_color_button)

        background_color_button = \
            self.main_window.widgets.background_color_button
        background_color_button.set_max_size(self.iconSize())
        self.add_widget(background_color_button)

        self.addSeparator()

        self.addAction(actions.copy_format)
        self.addAction(actions.paste_format)

        self.addSeparator()

        self.addWidget(self.get_manager_button())


class ChartTemplatesToolBar(ToolBarBase):
    """Toolbar for chart dialog for inserting template chart code"""

    tooltip_tpl = "Package {} required but not installed"

    def __init__(self, parent: QWidget):
        """
        :param parent: Parent widget, e.g. chart dialog window

        """

        super().__init__("Chart templates toolbar", parent)

        self.setObjectName("Chart templates toolbar")
        self._create_toolbar(parent.actions)

    def _create_toolbar(self, actions: ChartDialogActions):
        """Fills the chart dialog toolbar with QActions

        :param actions: Chart dialog actions

        """

        self.addAction(actions.chart_pie_1_1)
        self.addAction(actions.chart_ring_1_1)
        self.addAction(actions.chart_line_1_1)
        self.addAction(actions.chart_polar_1_1)
        self.addAction(actions.chart_area_1_1)
        self.addAction(actions.chart_column_1_1)
        self.addAction(actions.chart_column_1_2)
        self.addAction(actions.chart_bar_1_3)
        self.addAction(actions.chart_scatter_1_1)
        self.addAction(actions.chart_bubble_1_1)
        self.addAction(actions.chart_boxplot_2_2)
        self.addAction(actions.chart_histogram_1_1)
        self.addAction(actions.chart_histogram_1_4)
        self.addAction(actions.chart_scatterhist_1_1)
        self.addAction(actions.chart_matrix_1_1)
        self.addAction(actions.chart_contour_1_2)
        self.addAction(actions.chart_surface_2_1)

        self.addWidget(self.get_manager_button())


class RChartTemplatesToolBar(ToolBarBase):
    """Toolbar for chart dialog for inserting R template chart code

    Requires rpy2 with ggplot, lattice and graphics packages

    """

    tooltip_tpl = "R Package {} required but not installed"

    def __init__(self, parent: QWidget):
        """
        :param parent: Parent widget, e.g. chart dialog window

        """

        super().__init__("Chart templates toolbar", parent)

        if rpy2 is None:
            self.close()

        self.setObjectName("R Chart templates toolbar")
        self._create_toolbar(parent.actions)

    @staticmethod
    def is_r_package_installed(package_name):
        """True if the R package is installed

        :param package_name: Name of R package to checked

        """

        if rpy2 is None:
            return False

        try:
            importr(package_name)
        except RuntimeError:
            return False
        except PackageNotInstalledError:
            return False
        return True

    def _create_toolbar(self, actions: ChartDialogActions):
        """Fills the chart dialog toolbar with QActions

        :param actions: Chart dialog actions

        """

        self.addAction(actions.chart_r_graphics_barplot_1_1)
        self.addAction(actions.chart_r_ggplot2_geom_boxplot_1_2)
        self.addAction(actions.chart_r_ggplot2_geom_point_1_1)
        self.addAction(actions.chart_r_lattice_xyplot_1_1)
        self.addAction(actions.chart_r_ggplot2_geom_density2d_1_2)
        self.addAction(actions.chart_r_lattice_wireframe_2_1)
        self.addAction(actions.chart_plotnine_geom_bar_1_1)

        if not self.is_r_package_installed("graphics"):
            tooltip = self.tooltip_tpl.format("graphics")
            actions.chart_r_graphics_barplot_1_1.setEnabled(False)
            actions.chart_r_graphics_barplot_1_1.setToolTip(tooltip)

        if not self.is_r_package_installed("lattice"):
            tooltip = self.tooltip_tpl.format("lattice")
            actions.chart_r_lattice_xyplot_1_1.setEnabled(False)
            actions.chart_r_lattice_xyplot_1_1.setToolTip(tooltip)
            actions.chart_r_lattice_wireframe_2_1.setEnabled(False)
            actions.chart_r_lattice_wireframe_2_1.setToolTip(tooltip)

        if not self.is_r_package_installed("ggplot2"):
            tooltip = self.tooltip_tpl.format("ggplot2")
            actions.chart_r_ggplot2_geom_boxplot_1_2.setEnabled(False)
            actions.chart_r_ggplot2_geom_boxplot_1_2.setToolTip(tooltip)
            actions.chart_r_ggplot2_geom_point_1_1.setEnabled(False)
            actions.chart_r_ggplot2_geom_point_1_1.setToolTip(tooltip)
            actions.chart_r_ggplot2_geom_density2d_1_2.setEnabled(False)
            actions.chart_r_ggplot2_geom_density2d_1_2.setToolTip(tooltip)

        if plotnine is None:
            tooltip = self.tooltip_tpl.format("plotnine")
            actions.chart_plotnine_geom_bar_1_1.setEnabled(False)
            actions.chart_plotnine_geom_bar_1_1.setToolTip(tooltip)

        self.addWidget(self.get_manager_button())



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------


"""
test_grid
=========

Unit tests for grid.py

"""

from contextlib import contextmanager
from os.path import abspath, dirname, join
import sys

import pytest

from PyQt5.QtCore import QItemSelectionModel, QItemSelection
from PyQt5.QtWidgets import QApplication, QAbstractItemView
from PyQt5.QtGui import QFont, QColor


PYSPREADPATH = abspath(join(dirname(__file__) + "/.."))
LIBPATH = abspath(PYSPREADPATH + "/lib")


@contextmanager
def insert_path(path):
    sys.path.insert(0, path)
    yield
    sys.path.pop(0)


@contextmanager
def multi_selection_mode(grid):
    grid.clearSelection()
    old_selection_mode = grid.selectionMode()
    grid.setSelectionMode(QAbstractItemView.MultiSelection)
    yield
    grid.setSelectionMode(old_selection_mode)


with insert_path(PYSPREADPATH):
    from ..pyspread import MainWindow
    from ..commands import MakeButtonCell, RemoveButtonCell
    from ..lib.selection import Selection


app = QApplication.instance()
if app is None:
    app = QApplication([])
main_window = MainWindow()
zoom_levels = main_window.settings.zoom_levels


class TestGrid:
    """Unit tests for Grid in grid.py"""

    grid = main_window.grid
    cell_attributes = grid.model.code_array.cell_attributes

    param_test_row = [(0, 0), (1, 1), (100, 100), (1000, 0), (10000, 0),
                      (-1, 0)]

    @pytest.mark.parametrize("row, res", param_test_row)
    def test_row(self, row, res, monkeypatch):
        """Unit test for row getter and setter"""

        monkeypatch.setattr(self.grid, "row", row)
        assert self.grid.row == res

    param_test_column = [(0, 0), (1, 1), (100, 0), (1000, 0), (10000, 0),
                         (-1, 0)]

    @pytest.mark.parametrize("column, res", param_test_column)
    def test_column(self, column, res, monkeypatch):
        """Unit test for column getter and setter"""

        monkeypatch.setattr(self.grid, "column", column)
        assert self.grid.column == res

    param_test_table = [(0, 0), (1, 1), (3, 0), (-1, 0)]

    @pytest.mark.parametrize("table, res", param_test_table)
    def test_table(self, table, res, monkeypatch):
        """Unit test for table getter and setter"""

        monkeypatch.setattr(self.grid, "table", table)
        assert self.grid.table == res

    @pytest.mark.parametrize("row, row_res", param_test_row)
    @pytest.mark.parametrize("column, column_res", param_test_column)
    def test_current2(self, row, row_res, column, column_res, monkeypatch):
        """Unit test for current getter and setter with 2 parameters"""

        monkeypatch.setattr(self.grid, "current", (row, column))
        assert self.grid.current == (row_res, column_res, 0)

    @pytest.mark.parametrize("row, row_res", param_test_row)
    @pytest.mark.parametrize("column, column_res", param_test_column)
    @pytest.mark.parametrize("table, table_res", param_test_table)
    def test_current3(self, row, row_res, column, column_res, table, table_res,
                      monkeypatch):
        """Unit test for current getter and setter with 3 parameters"""

        monkeypatch.setattr(self.grid, "current", (row, column, table))
        assert self.grid.current == (row_res, column_res, table_res)

    def test_current_invalid(self):
        """Unit test for current getter and setter with invalid parameters"""

        with pytest.raises(ValueError):
            self.grid.current = 1, 2, 3, 4

    param_test_row_heights = [
        ({0: 23}, {0: 23}),
        ({1: 3}, {1: 3.0}),
        ({0: 23, 12: 200}, {0: 23, 12: 200}),
    ]

    @pytest.mark.parametrize("heights, heights_res", param_test_row_heights)
    def test_row_heights(self, heights, heights_res):
        """Unit test for row_heights"""

        for row in heights:
            self.grid.setRowHeight(row, heights[row])

        row_heights = dict(self.grid.row_heights)
        for row in heights_res:
            assert row_heights[row] == heights_res[row]

    param_test_column_widths = [
        ({0: 23}, {0: 23}),
        ({1: 3}, {1: 3.0}),
        ({0: 23, 12: 200}, {0: 23, 12: 200}),
    ]

    @pytest.mark.parametrize("widths, widths_res", param_test_column_widths)
    def test_column_widths(self, widths, widths_res):
        """Unit test for column_widths"""

        for column in widths:
            self.grid.setColumnWidth(column, widths[column])

        column_widths = dict(self.grid.column_widths)
        for column in widths_res:
            assert column_widths[column] == widths_res[column]

    param_test_selection_cells = [
        (((0, 0),), Selection([], [], [], [], [(0, 0)])),
        (((0, 0), (1, 0)), Selection([], [], [], [], [(0, 0), (1, 0)])),
        (((0, 0), (999, 0)), Selection([], [], [], [], [(0, 0), (999, 0)])),
        (((0, 0), (1, 0), (0, 1), (1, 1)),
         Selection([], [], [], [], [(0, 0), (1, 0), (0, 1), (1, 1)])),
    ]

    @pytest.mark.parametrize("cells, res", param_test_selection_cells)
    def test_selection_cells(self, cells, res):
        """Unit test for selection getter using cells"""

        with multi_selection_mode(self.grid):
            for cell in cells:
                idx = self.grid.model.index(*cell)
                self.grid.selectionModel().select(idx,
                                                  QItemSelectionModel.Select)
            assert self.grid.selection == res

    param_test_selection_blocks = [
        ((0, 0, 1, 1), Selection([(0, 0)], [(1, 1)], [], [], [])),
        ((1, 2, 14, 19), Selection([(1, 2)], [(14, 19)], [], [], [])),
    ]

    @pytest.mark.parametrize("block, res", param_test_selection_blocks)
    def test_selection_blocks(self, block, res):
        """Unit test for selection getter using blocks"""

        with multi_selection_mode(self.grid):
            top, left, bottom, right = block
            idx_tl = self.grid.model.index(top, left)
            idx_br = self.grid.model.index(bottom, right)
            item_selection = QItemSelection(idx_tl, idx_br)
            self.grid.selectionModel().select(item_selection,
                                              QItemSelectionModel.Select)
            assert self.grid.selection == res

    param_test_selection_rows = [
        ((0,), Selection([], [], [0], [], [])),
        ((1, 2), Selection([], [], [1, 2], [], [])),
    ]

    @pytest.mark.parametrize("rows, res", param_test_selection_rows)
    def test_selection_rows(self, rows, res):
        """Unit test for selection getter using rows"""

        with multi_selection_mode(self.grid):
            for row in rows:
                self.grid.selectRow(row)
            assert self.grid.selection == res

    param_test_selection_columns = [
        ((0,), Selection([], [], [], [0], [])),
        ((1, 2), Selection([], [], [], [1, 2], [])),
    ]

    @pytest.mark.parametrize("columns, res", param_test_selection_columns)
    def test_selection_columns(self, columns, res):
        """Unit test for selection getter using columns"""

        with multi_selection_mode(self.grid):
            for column in columns:
                self.grid.selectColumn(column)
            assert self.grid.selection == res

    param_test_zoom = [(1, 1), (2, 2), (8, 8), (0, 1), (100, 1), (-1, 1)]

    @pytest.mark.parametrize("zoom, zoom_res", param_test_zoom)
    def test_zoom(self, zoom, zoom_res, monkeypatch):
        """Unit test for zoom getter and setter"""

        monkeypatch.setattr(self.grid, "zoom", zoom)
        assert self.grid.zoom == zoom_res

    param_test_set_selection_mode = [
        (True, (0, 0, 0), (0, 0, 0), QAbstractItemView.NoEditTriggers),
        (False, (1, 2, 3), None,
         QAbstractItemView.DoubleClicked
         | QAbstractItemView.EditKeyPressed
         | QAbstractItemView.AnyKeyPressed),
    ]

    @pytest.mark.parametrize("on, current, start, edit_mode",
                             param_test_set_selection_mode)
    def test_set_selection_mode(self, on, current, start, edit_mode):
        """Unit test for set_selection_mode"""

        self.grid.set_selection_mode(False)
        self.grid.current = current
        self.grid.set_selection_mode(on)
        for grid in main_window.grids:
            assert grid.selection_mode == on
        assert self.grid.editTriggers() == edit_mode
        assert self.grid.current_selection_mode_start == start

    def test_focusInEvent(self):
        """Unit test for focusInEvent"""

        main_window.grids[0].setFocus()
        main_window.grids[1].setFocus()
        assert main_window._last_focused_grid == main_window.grids[0]

    def test_adjust_size(self):
        """Unit test for adjust_size"""

        w = self.grid.horizontalHeader().length() + \
            self.grid.verticalHeader().width()
        h = self.grid.verticalHeader().length() + \
            self.grid.horizontalHeader().height()

        self.grid.resize(200, 200)
        self.grid.adjust_size()
        assert self.grid.size().width() == w
        assert self.grid.size().height() == h

    param_test_selected_idx_to_str = [
        ([grid.model.createIndex(2, 4)], "(2, 4, 0)"),
        ([grid.model.createIndex(2, 4), grid.model.createIndex(3, 4)],
         "(2, 4, 0), (3, 4, 0)"),
    ]

    @pytest.mark.parametrize("sel_idx, res", param_test_selected_idx_to_str)
    def test_selected_idx_to_str(self, sel_idx, res):
        """Unit test for _selected_idx_to_str"""

        assert self.grid._selected_idx_to_str(sel_idx) == res

    def test_has_selection(self):
        """Unit test for has_selection"""

        assert not self.grid.has_selection()
        self.grid.selectRow(2)
        assert self.grid.has_selection()
        self.grid.clearSelection()
        assert not self.grid.has_selection()
        self.grid.selectColumn(2)
        self.grid.on_merge_pressed()
        self.grid.selectRow(2)
        assert self.grid.has_selection()
        self.grid.current = 2, 2, 0
        self.grid.selectColumn(2)
        assert self.grid.has_selection()

    def test_on_current_changed(self):
        """Unit test for on_current_changed"""

        main_window.entry_line.setPlainText("Test")
        self.grid.selection_mode_exiting = True
        self.grid.current = 0, 1, 0
        assert main_window.entry_line.toPlainText() == "Test"
        self.grid.selection_mode_exiting = False

        self.grid.selection_mode = True
        self.grid.current = 2, 1, 0
        assert main_window.entry_line.toPlainText() == "S[X + 2, Y + 0, Z]Test"
        self.grid.selection_mode = False
        main_window.entry_line.setPlainText("")

    def test_on_selection_changed(self):
        """Unit test for on_selection_changed"""

        main_window.settings.show_statusbar_sum = False
        self.grid.model.code_array[1, 0, 0] = "23"
        self.grid.model.code_array[2, 0, 0] = "2"
        self.grid.current = 0, 0, 0
        idx_tl = self.grid.model.index(0, 0)
        idx_br = self.grid.model.index(5, 0)
        item_selection = QItemSelection(idx_tl, idx_br)
        self.grid.selectionModel().select(item_selection,
                                          QItemSelectionModel.Select)
        assert main_window.statusBar().currentMessage() == ""

        self.grid.clearSelection()
        main_window.settings.show_statusbar_sum = True
        idx_tl = self.grid.model.index(0, 0)
        idx_br = self.grid.model.index(4, 0)
        item_selection = QItemSelection(idx_tl, idx_br)
        self.grid.selectionModel().select(item_selection,
                                          QItemSelectionModel.Select)
        assert main_window.statusBar().currentMessage() == \
               "Selection: 5 cells     Σ=25     max=23     min=2"

    def test_on_row_resized(self):
        """Unit test for on_row_resized"""

        self.grid.setRowHeight(2, 45)
        self.grid.setRowHeight(6, 45)

        row_heights = dict(self.grid.row_heights)
        assert row_heights[2] == 45
        assert row_heights[6] == 45

        self.grid.current = 3, 1, 0
        idx_tl = self.grid.model.index(3, 1)
        idx_br = self.grid.model.index(5, 2)
        item_selection = QItemSelection(idx_tl, idx_br)
        self.grid.selectionModel().select(item_selection,
                                          QItemSelectionModel.Select)

        self.grid.setRowHeight(3, 48)
        row_heights = dict(self.grid.row_heights)
        assert row_heights[2] == 45
        assert row_heights[3] == 48
        assert row_heights[4] == 48
        assert row_heights[5] == 48
        assert row_heights[6] == 45

        self.grid.clearSelection()

    def test_on_column_resized(self):
        """Unit test for on_column_resized"""

        self.grid.setColumnWidth(2, 45)
        self.grid.setColumnWidth(6, 45)

        col_widths = dict(self.grid.column_widths)
        assert col_widths[2] == 45
        assert col_widths[6] == 45

        self.grid.current = 1, 3, 0
        idx_tl = self.grid.model.index(1, 3)
        idx_br = self.grid.model.index(2, 5)
        item_selection = QItemSelection(idx_tl, idx_br)
        self.grid.selectionModel().select(item_selection,
                                          QItemSelectionModel.Select)

        self.grid.setColumnWidth(3, 48)
        col_widths = dict(self.grid.column_widths)
        assert col_widths[2] == 45
        assert col_widths[3] == 48
        assert col_widths[4] == 48
        assert col_widths[5] == 48
        assert col_widths[6] == 45

        self.grid.clearSelection()

    param_test_on_zoom_in = list(zip(zoom_levels[:-1], zoom_levels[1:]))
    param_test_on_zoom_in += [(max(zoom_levels), max(zoom_levels))]

    @pytest.mark.parametrize("zoom, res", param_test_on_zoom_in)
    def test_on_zoom_in(self, zoom, res):
        """Unit test for on_zoom_in"""

        for grid in main_window.grids:
            main_window._last_focused_grid = grid
            grid.zoom = zoom
            self.grid.on_zoom_in()
            assert grid.zoom == res
        main_window._last_focused_grid = self.grid

    param_test_on_zoom_out = list(zip(zoom_levels[1:], zoom_levels[:-1]))
    param_test_on_zoom_out += [(min(zoom_levels), min(zoom_levels))]

    @pytest.mark.parametrize("zoom, res", param_test_on_zoom_out)
    def test_on_zoom_out(self, zoom, res):
        """Unit test for on_zoom_out"""

        for grid in main_window.grids:
            main_window._last_focused_grid = grid
            grid.zoom = zoom
            self.grid.on_zoom_out()
            assert grid.zoom == res
        main_window._last_focused_grid = self.grid

    @pytest.mark.parametrize("zoom", zoom_levels)
    def test_on_zoom_1(self, zoom):
        """Unit test for on_zoom_1"""

        for grid in main_window.grids:
            main_window._last_focused_grid = grid
            grid.zoom = zoom
            grid.on_zoom_1()
            assert grid.zoom == 1.0
        main_window._last_focused_grid = self.grid

    def test_refresh_frozen_cell(self):
        """Unit test for _refresh_frozen_cell"""

        self.grid.current = 1, 0, 0
        self.grid.model.code_array[1, 0, 0] = "23"
        self.grid.on_freeze_pressed(True)
        self.grid.model.code_array[1, 0, 0] = "'Test'"
        assert self.grid.model.code_array.frozen_cache == {'(1, 0, 0)': 23}
        assert self.grid.model.code_array[1, 0, 0] == 23

        self.grid._refresh_frozen_cell((1, 0, 0))
        assert self.grid.model.code_array[1, 0, 0] == "Test"

        self.grid.on_freeze_pressed(False)

    def test_refresh_frozen_cells(self):
        """Unit test for refresh_frozen_cells"""

        self.grid.current = 1, 0, 0
        self.grid.model.code_array[1, 0, 0] = "23"
        self.grid.on_freeze_pressed(True)
        self.grid.current = 2, 0, 0
        self.grid.model.code_array[2, 0, 0] = "24"
        self.grid.on_freeze_pressed(True)

        assert self.grid.model.code_array[1, 0, 0] == 23
        assert self.grid.model.code_array[2, 0, 0] == 24

        self.grid.model.code_array[1, 0, 0] = "'Test1'"
        self.grid.model.code_array[2, 0, 0] = "'Test2'"

        assert self.grid.model.code_array[1, 0, 0] == 23
        assert self.grid.model.code_array[2, 0, 0] == 24

        self.grid.refresh_frozen_cells()

        assert self.grid.model.code_array[1, 0, 0] == "Test1"
        assert self.grid.model.code_array[2, 0, 0] == "Test2"

        self.grid.current = 1, 0, 0
        self.grid.on_freeze_pressed(False)
        self.grid.current = 2, 0, 0
        self.grid.on_freeze_pressed(False)

    def test_refresh_selected_frozen_cells(self):
        """Unit test for refresh_selected_frozen_cells"""

        self.grid.current = 1, 0, 0
        self.grid.model.code_array[1, 0, 0] = "23"
        self.grid.on_freeze_pressed(True)
        self.grid.current = 2, 0, 0
        self.grid.model.code_array[2, 0, 0] = "24"
        self.grid.on_freeze_pressed(True)

        assert self.grid.model.code_array[1, 0, 0] == 23
        assert self.grid.model.code_array[2, 0, 0] == 24

        self.grid.model.code_array[1, 0, 0] = "'Test1'"
        self.grid.model.code_array[2, 0, 0] = "'Test2'"

        assert self.grid.model.code_array[1, 0, 0] == 23
        assert self.grid.model.code_array[2, 0, 0] == 24

        self.grid.selectRow(1)
        self.grid.refresh_selected_frozen_cells()

        assert self.grid.model.code_array[1, 0, 0] == "Test1"
        assert self.grid.model.code_array[2, 0, 0] == 24

        self.grid.current = 1, 0, 0
        self.grid.on_freeze_pressed(False)
        self.grid.current = 2, 0, 0
        self.grid.on_freeze_pressed(False)

    def test_on_show_frozen_pressed(self):
        """Unit test for on_show_frozen_pressed"""

        self.grid.on_show_frozen_pressed(True)
        assert main_window.settings.show_frozen
        self.grid.on_show_frozen_pressed(False)
        assert not main_window.settings.show_frozen

    def test_on_font_size(self):
        """Unit test for on_font_size"""

        self.grid.selectRow(2)

        main_window.widgets.font_size_combo.size = 14
        assert main_window.widgets.font_size_combo.size == 14
        self.grid.on_font_size()
        assert self.cell_attributes[(2, 0, 0)]["pointsize"] == 14

        self.grid.clearSelection()

    def test_on_bold_pressed(self):
        """Unit test for on_bold_pressed"""

        self.grid.selectRow(2)

        self.grid.on_bold_pressed(True)
        assert self.cell_attributes[(2, 0, 0)]["fontweight"] == QFont.Bold
        self.grid.on_bold_pressed(False)
        assert self.cell_attributes[(2, 0, 0)]["fontweight"] == QFont.Normal

        self.grid.clearSelection()

    def test_on_italics_pressed(self):
        """Unit test for on_italics_pressed"""

        self.grid.selectRow(2)

        self.grid.on_italics_pressed(True)
        assert self.cell_attributes[(2, 0, 0)]["fontstyle"] \
            == QFont.StyleItalic
        self.grid.on_italics_pressed(False)
        assert self.cell_attributes[(2, 0, 0)]["fontstyle"] \
            == QFont.StyleNormal

        self.grid.clearSelection()

    def test_on_underline_pressed(self):
        """Unit test for on_underline_pressed"""

        self.grid.selectRow(2)

        self.grid.on_underline_pressed(True)
        assert self.cell_attributes[(2, 0, 0)]["underline"]
        self.grid.on_underline_pressed(False)
        assert not self.cell_attributes[(2, 0, 0)]["underline"]

        self.grid.clearSelection()

    def test_on_strikethrough_pressed(self):
        """Unit test for on_strikethrough_pressed"""

        self.grid.selectRow(2)

        self.grid.on_strikethrough_pressed(True)
        assert self.cell_attributes[(2, 0, 0)]["strikethrough"]
        self.grid.on_strikethrough_pressed(False)
        assert not self.cell_attributes[(2, 0, 0)]["strikethrough"]

        self.grid.clearSelection()

    def test_on_text_renderer_pressed(self):
        """Unit test for on_text_renderer_pressed"""

        self.grid.selectRow(2)

        self.grid.on_text_renderer_pressed()
        assert self.cell_attributes[(2, 0, 0)]["renderer"] == "text"
        self.grid.on_text_renderer_pressed()
        assert self.cell_attributes[(2, 0, 0)]["renderer"] == "text"

        self.grid.clearSelection()

    def test_on_image_renderer_pressed(self):
        """Unit test for on_image_renderer_pressed"""

        self.grid.selectRow(2)

        self.grid.on_image_renderer_pressed()
        assert self.cell_attributes[(2, 0, 0)]["renderer"] == "image"
        self.grid.on_text_renderer_pressed()
        assert self.cell_attributes[(2, 0, 0)]["renderer"] == "text"

        self.grid.clearSelection()

    def test_on_markup_renderer_pressed(self):
        """Unit test for on_markup_renderer_pressed"""

        self.grid.selectRow(2)

        self.grid.on_markup_renderer_pressed()
        assert self.cell_attributes[(2, 0, 0)]["renderer"] == "markup"
        self.grid.on_text_renderer_pressed()
        assert self.cell_attributes[(2, 0, 0)]["renderer"] == "text"

        self.grid.clearSelection()

    def test_on_matplotlib_renderer_pressed(self):
        """Unit test for on_matplotlib_renderer_pressed"""

        self.grid.selectRow(2)

        self.grid.on_matplotlib_renderer_pressed()
        assert self.cell_attributes[(2, 0, 0)]["renderer"] == "matplotlib"
        self.grid.on_text_renderer_pressed()
        assert self.cell_attributes[(2, 0, 0)]["renderer"] == "text"

        self.grid.clearSelection()

    def test_on_lock_pressed(self):
        """Unit test for on_lock_pressed"""

        self.grid.selectRow(2)

        self.grid.on_lock_pressed(True)
        assert self.cell_attributes[(2, 0, 0)]["locked"]
        self.grid.on_lock_pressed(False)
        assert not self.cell_attributes[(2, 0, 0)]["locked"]

        self.grid.clearSelection()

    def test_on_rotate_0(self):
        """Unit test for on_rotate_0"""

        self.grid.selectRow(2)

        self.grid.on_rotate_0()
        assert self.cell_attributes[(2, 0, 0)]["angle"] == 0.0

        self.grid.clearSelection()

    def test_on_rotate_90(self):
        """Unit test for on_rotate_90"""

        self.grid.selectRow(2)

        self.grid.on_rotate_90()
        assert self.cell_attributes[(2, 0, 0)]["angle"] == 90.0
        self.grid.on_rotate_0()
        assert self.cell_attributes[(2, 0, 0)]["angle"] == 0.0

        self.grid.clearSelection()

    def test_on_rotate_180(self):
        """Unit test for on_rotate_180"""

        self.grid.selectRow(2)

        self.grid.on_rotate_180()
        assert self.cell_attributes[(2, 0, 0)]["angle"] == 180.0
        self.grid.on_rotate_0()
        assert self.cell_attributes[(2, 0, 0)]["angle"] == 0.0

        self.grid.clearSelection()

    def test_on_rotate_270(self):
        """Unit test for on_rotate_270"""

        self.grid.selectRow(2)

        self.grid.on_rotate_270()
        assert self.cell_attributes[(2, 0, 0)]["angle"] == 270.0
        self.grid.on_rotate_0()
        assert self.cell_attributes[(2, 0, 0)]["angle"] == 0.0

        self.grid.clearSelection()

    def test_on_justify_left(self):
        """Unit test for on_justify_left"""

        self.grid.selectRow(2)

        self.grid.on_justify_left()
        assert self.cell_attributes[(2, 0, 0)]["justification"] \
            == "justify_left"

        self.grid.clearSelection()

    def test_on_justify_fill(self):
        """Unit test for on_justify_fill"""

        self.grid.selectRow(2)

        self.grid.on_justify_fill()
        assert self.cell_attributes[(2, 0, 0)]["justification"] \
            == "justify_fill"
        self.grid.on_justify_left()
        assert self.cell_attributes[(2, 0, 0)]["justification"] \
            == "justify_left"

        self.grid.clearSelection()

    def test_on_justify_center(self):
        """Unit test for on_justify_center"""

        self.grid.selectRow(2)

        self.grid.on_justify_center()
        assert self.cell_attributes[(2, 0, 0)]["justification"] \
            == "justify_center"
        self.grid.on_justify_left()
        assert self.cell_attributes[(2, 0, 0)]["justification"] \
            == "justify_left"

        self.grid.clearSelection()

    def test_on_justify_right(self):
        """Unit test for on_justify_right"""

        self.grid.selectRow(2)

        self.grid.on_justify_right()
        assert self.cell_attributes[(2, 0, 0)]["justification"] \
            == "justify_right"
        self.grid.on_justify_left()
        assert self.cell_attributes[(2, 0, 0)]["justification"] \
            == "justify_left"

        self.grid.clearSelection()

    def test_on_align_top(self):
        """Unit test for on_align_top"""

        self.grid.selectRow(2)

        self.grid.on_align_top()
        assert self.cell_attributes[(2, 0, 0)]["vertical_align"] == "align_top"

    def test_on_align_middle(self):
        """Unit test for on_align_middle"""

        self.grid.selectRow(2)

        self.grid.on_align_middle()
        assert self.cell_attributes[(2, 0, 0)]["vertical_align"] \
            == "align_center"
        self.grid.on_align_top()
        assert self.cell_attributes[(2, 0, 0)]["vertical_align"] == "align_top"

    def test_on_align_bottom(self):
        """Unit test for on_align_bottom"""

        self.grid.selectRow(2)

        self.grid.on_align_bottom()
        assert self.cell_attributes[(2, 0, 0)]["vertical_align"] \
            == "align_bottom"
        self.grid.on_align_top()
        assert self.cell_attributes[(2, 0, 0)]["vertical_align"] == "align_top"

    def test_on_text_color(self):
        """Unit test for on_text_color"""

        self.grid.selectRow(2)
        main_window.widgets.text_color_button.color = QColor(100, 100, 50)

        self.grid.on_text_color()

        assert self.cell_attributes[(2, 0, 0)]["textcolor"] \
            == (100, 100, 50, 255)

        self.grid.clearSelection()

    def test_on_line_color(self):
        """Unit test for on_line_color"""

        self.grid.selectRow(2)
        main_window.widgets.line_color_button.color = QColor(100, 100, 50)

        self.grid.on_line_color()

        assert self.cell_attributes[(2, 0, 0)]["bordercolor_bottom"] \
            == (100, 100, 50, 255)

        assert self.cell_attributes[(2, 99, 0)]["bordercolor_right"] \
            == (100, 100, 50, 255)

        self.grid.clearSelection()

    def test_on_background_color(self):
        """Unit test for on_background_color"""

        self.grid.selectRow(2)
        main_window.widgets.background_color_button.color = QColor(100, 10, 5)

        self.grid.on_background_color()

        assert self.cell_attributes[(2, 0, 0)]["bgcolor"] == (100, 10, 5, 255)

        self.grid.clearSelection()

    def test_update_cell_spans(self):
        """Unit test for update_cell_spans"""

        self.cell_attributes.clear()
        self.grid.clearSelection()
        self.grid.selectRow(2)
        self.grid.on_merge_pressed()
        self.grid.table = 1
        self.grid.update_cell_spans()
        self.grid.table = 0
        self.grid.update_cell_spans()

        assert self.grid.columnSpan(2, 0) == 100

        self.grid.selectRow(2)
        self.grid.on_merge_pressed()
        self.grid.clearSelection()
        self.grid.selectColumn(1)
        self.grid.on_merge_pressed()
        self.grid.update_cell_spans()

        assert self.grid.rowSpan(0, 1) == 1000

        self.grid.selectColumn(1)
        self.grid.on_merge_pressed()
        self.grid.update_cell_spans()

    def test_update_index_widgets(self):
        """Unit test for update_index_widgets"""

        self.grid.current = 2, 2, 0
        description_tpl = "Make cell {} a button cell"
        description = description_tpl.format(self.grid.current)
        command = MakeButtonCell(self.grid, "TestButton",
                                 self.grid.currentIndex(), description)
        main_window.undo_stack.push(command)

        self.grid.update_index_widgets()
        assert self.grid.widget_indices

        description_tpl = "Make cell {} a non-button cell"
        description = description_tpl.format(self.grid.current)
        command = RemoveButtonCell(self.grid, self.grid.currentIndex(),
                                   description)
        main_window.undo_stack.push(command)

        self.grid.update_index_widgets()
        assert not self.grid.widget_indices

    def test_on_freeze_pressed(self):
        """Unit test for on_freeze_pressed"""

        self.grid.current = 1, 0, 0
        self.grid.model.code_array[1, 0, 0] = "23"
        self.grid.on_freeze_pressed(False)
        assert not self.cell_attributes[self.grid.current]["frozen"]
        self.grid.on_freeze_pressed(True)
        assert self.cell_attributes[self.grid.current]["frozen"]
        self.grid.on_freeze_pressed(False)
        assert not self.cell_attributes[self.grid.current]["frozen"]

    def test_on_merge_pressed(self):
        """Unit test for on_merge_pressed"""

        self.grid.clearSelection()
        self.grid.current = 1, 0, 0
        self.grid.on_merge_pressed()

        assert not self.cell_attributes[self.grid.current]["merge_area"]

    def test_on_quote(self):
        """Unit test for on_quote"""

        self.grid.model.code_array[1, 0, 0] = "42"

        self.grid.clearSelection()
        self.grid.selectRow(1)

        self.grid.on_quote()
        assert self.grid.model.code_array((1, 0, 0)) == "'42'"

    def test_is_row_data_discarded(self):
        """Unit test for is_row_data_discarded"""

        self.grid.model.code_array[998, 0, 0] = "Edge data"

        assert not self.grid.is_row_data_discarded(0)
        assert not self.grid.is_row_data_discarded(1)
        assert self.grid.is_row_data_discarded(2)

        self.grid.model.code_array[998, 0, 0] = None

    def test_is_column_data_discarded(self):
        """Unit test for is_column_data_discarded"""

        self.grid.model.code_array[0, 98, 0] = "Edge data"

        assert not self.grid.is_column_data_discarded(0)
        assert not self.grid.is_column_data_discarded(1)
        assert self.grid.is_column_data_discarded(2)

        self.grid.model.code_array[0, 98, 0] = None

    def test_is_table_data_discarded(self):
        """Unit test for is_table_data_discarded"""

        self.grid.model.code_array[0, 0, 1] = "Edge data"

        assert not self.grid.is_table_data_discarded(0)
        assert not self.grid.is_table_data_discarded(1)
        assert self.grid.is_table_data_discarded(2)

        self.grid.model.code_array[0, 0, 1] = None

    def test_on_insert_rows(self):
        """Unit test for on_insert_rows"""

        self.grid.clearSelection()
        self.grid.model.reset()

        self.current = 0, 0, 0
        self.grid.model.code_array[1, 0, 0] = "'Test data'"
        self.grid.selectRow(0)
        self.grid.on_insert_rows()
        assert self.grid.model.code_array[1, 0, 0] is None
        assert self.grid.model.code_array[2, 0, 0] == "Test data"

        self.grid.clearSelection()
        self.grid.selectRow(1)
        self.grid.on_insert_rows()
        assert self.grid.model.code_array[3, 0, 0] == "Test data"

    def test_on_delete_rows(self):
        """Unit test for on_delete_rows"""

        self.grid.clearSelection()
        self.grid.model.reset()

        self.current = 0, 0, 0
        self.grid.model.code_array[1, 0, 0] = "'Test data'"
        self.grid.selectRow(0)
        self.grid.on_delete_rows()
        assert self.grid.model.code_array[1, 0, 0] is None
        assert self.grid.model.code_array[0, 0, 0] == "Test data"

    def test_on_insert_columns(self):
        """Unit test for on_insert_columns"""

        self.grid.clearSelection()
        self.grid.model.reset()

        self.current = 0, 0, 0
        self.grid.model.code_array[0, 1, 0] = "'Test data'"
        self.grid.selectColumn(0)
        self.grid.on_insert_columns()
        assert self.grid.model.code_array[0, 1, 0] is None
        assert self.grid.model.code_array[0, 2, 0] == "Test data"

        self.grid.clearSelection()
        self.grid.selectColumn(1)
        self.grid.on_insert_columns()
        assert self.grid.model.code_array[0, 3, 0] == "Test data"

    def test_on_delete_columns(self):
        """Unit test for on_delete_columns"""

        self.grid.clearSelection()
        self.grid.model.reset()

        self.current = 0, 0, 0
        self.grid.model.code_array[0, 1, 0] = "'Test data'"
        self.grid.selectColumn(0)
        self.grid.on_delete_columns()
        assert self.grid.model.code_array[0, 1, 0] is None
        assert self.grid.model.code_array[0, 0, 0] == "Test data"

    def test_on_insert_table(self):
        """Unit test for on_insert_table"""

        self.grid.clearSelection()
        self.grid.model.reset()

        self.current = 0, 0, 0
        self.grid.model.code_array[0, 0, 1] = "'Test data'"
        self.grid.on_insert_table()
        assert self.grid.model.code_array[0, 0, 1] is None
        assert self.grid.model.code_array[0, 0, 2] == "Test data"

    def test_on_delete_table(self):
        """Unit test for on_delete_table"""

        self.grid.clearSelection()
        self.grid.model.reset()

        self.current = 0, 0, 0
        self.grid.model.code_array[0, 0, 1] = "'Test data'"
        self.grid.on_delete_table()
        assert self.grid.model.code_array[0, 0, 1] is None
        assert self.grid.model.code_array[0, 0, 0] == "Test data"


class TestGridHeaderView:
    """Unit tests for GridHeaderView in grid.py"""

    def test_sectionSizeHint(self):
        """Unit test for sectionSizeHint"""

        hghview = main_window.grid.horizontalHeader()
        size1 = hghview.sectionSizeHint(0)
        main_window.grid.zoom = 2.0
        assert hghview.sectionSizeHint(0) == 2 * size1


class TestGridTableModel:
    """Unit tests for GridTableModel in grid.py"""

    model = main_window.grid.model

    param_test_shape = [
        ((1, 1, 1), (1, 1, 1), None),
        ((0, 0, 0), (1000, 100, 3), ValueError),
        ((9999999999, 0, 0), (1000, 100, 3), ValueError),
        ((1000000, 10000, 10), (1000000, 10000, 10), None),
        ((1000, 100, 3), (1000, 100, 3), None),
    ]

    @pytest.mark.parametrize("shape, res, error", param_test_shape)
    def test_shape(self, shape, res, error, monkeypatch):
        """Unit test for shape getter and setter"""

        try:
            monkeypatch.setattr(self.model, "shape", shape)
        except ValueError:
            assert error == ValueError
        else:
            assert error is None
        assert self.model.shape == res

    param_test_code = [
        (0, 0, "", None),
        (0, 0, "None", "None"),
        (0, 0, "2+6", "2+6"),
        (1, 1, "test", "test"),
    ]

    @pytest.mark.parametrize("row, column, code, res", param_test_code)
    def test_code(self, row, column, code, res):
        """Unit test for code"""

        class Index:
            def __init__(self, row: int, column: int):
                self._row = row
                self._column = column

            def row(self) -> int:
                return self._row

            def column(self) -> int:
                return self._column

        self.model.code_array[(row, column, 0)] = code
        index = Index(row, column)
        assert self.model.code(index) == res

    param_test_insertRows = [
        (0, 5, (0, 0, 0), "0", (5, 0, 0), "0"),
        (0, 5, (0, 0, 0), "0", (0, 0, 0), None),
        (0, 0, (0, 0, 0), "0", (0, 0, 0), "0"),
        (3, 5, (0, 0, 0), "0", (0, 0, 0), "0"),
        (0, 500, (0, 0, 0), "0", (500, 0, 0), "0"),
    ]

    @pytest.mark.parametrize("row, count, key, code, reskey, res",
                             param_test_insertRows)
    def test_insertRows(self, row, count, key, code, reskey, res):
        """Unit test for insertRows"""

        self.model.code_array[key] = code
        self.model.insertRows(row, count)
        assert self.model.code_array(reskey) == res

    param_test_removeRows = [
        (0, 5, (5, 0, 0), "0", (0, 0, 0), "0"),
        (0, 5, (0, 0, 0), "0", (0, 0, 0), None),
        (0, 0, (0, 0, 0), "0", (0, 0, 0), "0"),
        (3, 5, (0, 0, 0), "0", (0, 0, 0), "0"),
        (0, 499, (500, 0, 0), "0", (1, 0, 0), "0"),
    ]

    @pytest.mark.parametrize("row, count, key, code, reskey, res",
                             param_test_removeRows)
    def test_removeRows(self, row, count, key, code, reskey, res):
        """Unit test for removeRows"""

        self.model.code_array[key] = code
        self.model.removeRows(row, count)
        assert self.model.code_array(reskey) == res

    param_test_insertColumns = [
        (0, 5, (0, 0, 0), "0", (0, 5, 0), "0"),
        (0, 5, (0, 0, 0), "0", (0, 0, 0), None),
        (0, 0, (0, 0, 0), "0", (0, 0, 0), "0"),
        (3, 5, (0, 0, 0), "0", (0, 0, 0), "0"),
        (0, 50, (0, 0, 0), "0", (0, 50, 0), "0"),
    ]

    @pytest.mark.parametrize("column, count, key, code, reskey, res",
                             param_test_insertColumns)
    def test_insertColumns(self, column, count, key, code, reskey, res):
        """Unit test for insertColumns"""

        self.model.code_array[key] = code
        self.model.insertColumns(column, count)
        assert self.model.code_array(reskey) == res

    param_test_removeColumns = [
        (0, 2, (0, 2, 0), "0", (0, 0, 0), "0"),
        (0, 2, (0, 0, 0), "0", (0, 0, 0), None),
        (0, 0, (0, 0, 0), "0", (0, 0, 0), "0"),
        (3, 1, (0, 0, 0), "0", (0, 0, 0), "0"),
    ]

    @pytest.mark.parametrize("column, count, key, code, reskey, res",
                             param_test_removeColumns)
    def test_removeColumns(self, column, count, key, code, reskey, res):
        """Unit test for removeColumns"""

        self.model.code_array[key] = code
        self.model.removeColumns(column, count)
        assert self.model.code_array(reskey) == res

    param_test_insertTable = [
        (0, (0, 0, 0), "0", (0, 0, 1), "0"),
        (0, (0, 0, 0), "0", (0, 0, 0), None),
        (2, (0, 0, 0), "0", (0, 0, 0), "0"),
    ]

    @pytest.mark.parametrize("table, key, code, reskey, res",
                             param_test_insertTable)
    def test_insertTable(self, table, key, code, reskey, res):
        """Unit test for insertTable"""

        self.model.code_array[key] = code
        self.model.insertTable(table)
        assert self.model.code_array(reskey) == res

    param_test_removeTable = [
        (0, (0, 0, 1), "0", (0, 0, 0), "0"),
        (0, (0, 0, 0), "0", (0, 0, 0), None),
        (1, (0, 0, 2), "0", (0, 0, 1), "0"),
    ]

    @pytest.mark.parametrize("table, key, code, reskey, res",
                             param_test_removeTable)
    def test_removeTable(self, table, key, code, reskey, res):
        """Unit test for removeTable"""

        self.model.code_array[key] = code
        self.model.removeTable(table)
        assert self.model.code_array(reskey) == res

    def test_reset(self):
        """Unit test for reset"""

        self.model.reset()

        assert not self.model.code_array.dict_grid
        assert not self.model.code_array.dict_grid.cell_attributes
        assert not self.model.code_array.row_heights
        assert not self.model.code_array.col_widths
        assert not self.model.code_array.macros
        assert not self.model.code_array.result_cache


class TestGridCellDelegate:
    """Unit tests for GridCellDelegate in grid.py"""


class TestTableChoice:
    """Unit tests for TableChoice in grid.py"""



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

 * :class:`Action` is a quick one liner way to create `QAction`
 * :class:`MainWindowActions`
 * :class:`ChartDialogActions`
 * :class:`SpellTextEditActions`

"""

from typing import Callable, List

from PyQt5.QtWidgets import QAction, QActionGroup, QWidget
from PyQt5.QtGui import QKeySequence, QIcon

try:
    import matplotlib.figure as matplotlib_figure
except ImportError:
    matplotlib_figure = None

try:
    import enchant
except ImportError:
    enchant = None

try:
    from pyspread.icons import Icon
    from pyspread.lib.attrdict import AttrDict
except ImportError:
    from icons import Icon
    from lib.attrdict import AttrDict


class Action(QAction):
    """A convenience class for creating a `QAction`

    .. Note: Parameter order has changed comparing with QAction

    """

    def __init__(self, parent: QWidget, label: str, *callbacks: List[Callable],
                 icon: QIcon = None, shortcut: str = None,
                 statustip: str = None, checkable: bool = False,
                 role: QAction.MenuRole = None):
        """

        :param parent: The parent object, normally :class:`pyspread.MainWindow`
        :param label: The text to appear
        :param callbacks: the callback functions
        :param icon: the :class:`icons.Icon`
        :param shortcut: The magic kestrokes if ant
        :param statustip: The popup message
        :param checkable: Has a checkbox
        :param role: Menu role for action for macOS

        """
        if icon is None:
            super().__init__(label, parent, checkable=checkable)
        else:
            super().__init__(icon, label, parent, checkable=checkable)

        if shortcut is not None:
            self.setShortcut(shortcut)

        if statustip is not None:
            self.setStatusTip(statustip)

        if role is not None:
            self.setMenuRole(role)

        for connect in callbacks:
            self.triggered.connect(connect)


class MainWindowActions(AttrDict):
    """Holds all QActions for the main window"""

    def __init__(self, parent: QWidget, shortcuts: bool = True):
        """
        :param parent: The parent object, normally :class:`pyspread.MainWindow`
        :param shortcuts: Enable shortcuts for actions

        """

        super().__init__()
        self.parent = parent
        self.shortcuts = shortcuts

        self.create_file_actions()
        self.create_edit_actions()
        self.create_view_actions()
        self.create_format_actions()
        self.create_macro_actions()
        self.create_help_actions()

        self.disable_unavailable()

    def create_file_actions(self):
        """actions for File menu"""

        self.new = Action(self.parent, "&New",
                          self.parent.workflows.file_new,
                          icon=Icon.new,
                          shortcut='Ctrl+n' if self.shortcuts else "",
                          statustip='Create a new, empty spreadsheet')

        self.open = Action(self.parent, "&Open",
                           self.parent.workflows.file_open,
                           icon=Icon.open,
                           statustip='Open spreadsheet from file')

        self.save = Action(self.parent, "&Save",
                           self.parent.workflows.file_save,
                           icon=Icon.save,
                           shortcut='Ctrl+s' if self.shortcuts else "",
                           statustip='Save spreadsheet')

        self.save_as = Action(
            self.parent, "Save &As",
            self.parent.workflows.file_save_as,
            icon=Icon.save_as,
            shortcut='Shift+Ctrl+s' if self.shortcuts else "",
            statustip='Save spreadsheet to a new file')

        self.imprt = Action(self.parent, "&Import",
                            self.parent.workflows.file_import,
                            icon=Icon.imprt,
                            statustip='Import a file and paste it into the '
                                      'current grid')

        self.export = Action(self.parent, "&Export",
                             self.parent.workflows.file_export,
                             icon=Icon.export,
                             statustip="Export selection to a file")

        self.approve = Action(self.parent, "&Approve file",
                              self.parent.on_approve,
                              icon=Icon.approve,
                              statustip='Approve, unfreeze and sign the '
                                        'current file')

        self.clear_globals = Action(self.parent, "&Clear globals",
                                    self.parent.on_clear_globals,
                                    icon=Icon.clear_globals,
                                    statustip='Deletes global variables '
                                              'and reloads base modules')

        self.print_preview = Action(self.parent, "Print preview",
                                    self.parent.on_preview,
                                    icon=Icon.print_preview,
                                    statustip='Print preview')

        self.print = Action(self.parent, "Print", self.parent.on_print,
                            icon=Icon.print,
                            shortcut='Ctrl+p' if self.shortcuts else "",
                            statustip='Print current spreadsheet')

        self.preferences = Action(self.parent, "Preferences...",
                                  self.parent.on_preferences,
                                  icon=Icon.preferences,
                                  statustip='Pyspread setup parameters',
                                  role=QAction.PreferencesRole)

        self.quit = Action(self.parent, "&Quit", self.parent.closeEvent,
                           icon=Icon.quit,
                           shortcut='Ctrl+Q' if self.shortcuts else "",
                           statustip='Exit pyspread', role=QAction.QuitRole)

    def create_edit_actions(self):
        """actions for Edit menu"""

        self.undo = Action(self.parent, "&Undo",
                           self.parent.on_undo,
                           icon=Icon.undo,
                           shortcut='Ctrl+z' if self.shortcuts else "",
                           statustip='Undo last step')

        self.redo = Action(self.parent, "&Redo",
                           self.parent.on_redo,
                           icon=Icon.redo,
                           shortcut='Shift+Ctrl+z' if self.shortcuts else "",
                           statustip='Redo last undone step')

        self.cut = Action(self.parent, "Cut",
                          self.parent.workflows.edit_cut,
                          icon=Icon.cut,
                          shortcut='Ctrl+x' if self.shortcuts else "",
                          statustip='Cut cell to the clipboard')

        self.copy = Action(self.parent, "&Copy",
                           self.parent.workflows.edit_copy,
                           icon=Icon.copy,
                           shortcut='Ctrl+c' if self.shortcuts else "",
                           statustip='Copy the input strings of the cells '
                                     'to the clipboard')

        self.copy_results = \
            Action(self.parent, "Copy results",
                   self.parent.workflows.edit_copy_results,
                   icon=Icon.copy_results,
                   shortcut='Shift+Ctrl+c' if self.shortcuts else "",
                   statustip='Copy the result strings of the cells to the '
                             'clipboard')

        self.paste = Action(self.parent, "&Paste",
                            self.parent.workflows.edit_paste,
                            icon=Icon.paste,
                            shortcut='Ctrl+v' if self.shortcuts else "",
                            statustip='Paste cells from the clipboard')

        self.paste_as = Action(
            self.parent, "Paste as...",
            self.parent.workflows.edit_paste_as,
            icon=Icon.paste_as,
            shortcut='Shift+Ctrl+v' if self.shortcuts else "",
            statustip='Transform clipboard and paste results')

        self.find = Action(self.parent, "&Find...",
                           self.parent.workflows.edit_find,
                           icon=Icon.find,
                           shortcut='Ctrl+f' if self.shortcuts else "",
                           statustip='Find dialog')

        self.find_next = Action(self.parent, "&Find next",
                                self.parent.workflows.edit_find_next,
                                icon=Icon.find_next,
                                shortcut='F3' if self.shortcuts else "",
                                statustip='Find next matching cell')

        self.replace = Action(
            self.parent, "&Replace...",
            self.parent.workflows.edit_replace,
            icon=Icon.replace,
            shortcut='Shift+Ctrl+f' if self.shortcuts else "",
            statustip='Replace sub-strings in cells')

        self.sort_ascending = Action(
            self.parent, "Sort ascending",
            self.parent.workflows.edit_sort_ascending,
            icon=Icon.sort_ascending,
            statustip='Sort selected cells. The sort order is ascending and '
                      'follows the current column.')

        self.sort_descending = Action(
            self.parent, "Sort descending",
            self.parent.workflows.edit_sort_descending,
            icon=Icon.sort_descending,
            statustip='Sort selected cells. The sort order is descending and '
                      'follows the current column.')

        self.toggle_selection_mode = Action(
            self.parent, "Selection mode",
            self.parent.grid.set_selection_mode,
            icon=Icon.selection_mode, checkable=True,
            shortcut='Ins',
            statustip='Enter/leave selection mode')

        self.quote = Action(self.parent, "&Quote",
                            self.parent.grid.on_quote,
                            icon=Icon.quote,
                            shortcut='Ctrl+Return' if self.shortcuts else "",
                            statustip="Convert cells' code to strings by "
                                      "addding quotes")

        self.insert_rows = Action(self.parent, "Insert rows",
                                  self.parent.grid.on_insert_rows,
                                  icon=Icon.insert_row,
                                  statustip='Insert max(1, no. selected '
                                            'rows) rows at cursor')

        self.insert_columns = Action(self.parent, "Insert columns",
                                     self.parent.grid.on_insert_columns,
                                     icon=Icon.insert_column,
                                     statustip='Insert max(1, no. selected '
                                               'columns) columns at cursor')

        self.insert_table = Action(self.parent, "Insert table",
                                   self.parent.grid.on_insert_table,
                                   icon=Icon.insert_table,
                                   statustip='Insert table before current '
                                             'table')

        self.delete_rows = Action(self.parent, "Delete rows",
                                  self.parent.grid.on_delete_rows,
                                  icon=Icon.delete_row,
                                  statustip='Delete max(1, no. selected '
                                            'rows) rows at cursor')

        self.delete_columns = Action(self.parent, "Delete columns",
                                     self.parent.grid.on_delete_columns,
                                     icon=Icon.delete_column,
                                     statustip='Delete max(1, no. selected '
                                               'columns) columns at cursor')

        self.delete_table = Action(self.parent, "Delete table",
                                   self.parent.grid.on_delete_table,
                                   icon=Icon.delete_table,
                                   statustip='Delete current table')

        self.resize_grid = Action(self.parent, "Resize grid",
                                  self.parent.workflows.edit_resize,
                                  icon=Icon.resize_grid,
                                  statustip='Resizes the current grid')

    def create_view_actions(self):
        """actions for View menu"""

        self.fullscreen = Action(self.parent, "Fullscreen",
                                 self.parent.on_fullscreen,
                                 icon=Icon.fullscreen,
                                 shortcut='F11' if self.shortcuts else "",
                                 statustip='Show grid in fullscreen mode '
                                           '(press <F11> to leave)')

        self.toggle_main_toolbar = Action(self.parent, "Main toolbar",
                                          self.parent.on_toggle_main_toolbar,
                                          checkable=True,
                                          statustip='Show/hide the main '
                                                    'toolbar')

        self.toggle_macro_toolbar = Action(self.parent, "Macro toolbar",
                                           self.parent.on_toggle_macro_toolbar,
                                           checkable=True,
                                           statustip='Show/hide the macro '
                                                     'toolbar')

        self.toggle_format_toolbar = \
            Action(self.parent, "Format toolbar",
                   self.parent.on_toggle_format_toolbar,
                   checkable=True,
                   statustip='Show/hide the format toolbar')

        self.toggle_find_toolbar = Action(self.parent, "Find toolbar",
                                          self.parent.on_toggle_find_toolbar,
                                          checkable=True,
                                          statustip='Show/hide the find '
                                                    'toolbar')

        self.toggle_entry_line_dock = Action(
            self.parent, "Entry line", self.parent.on_toggle_entry_line_dock,
            checkable=True, statustip='Show/hide the entry line')

        self.toggle_macro_dock = Action(
            self.parent, "Macro panel", self.parent.on_toggle_macro_dock,
            checkable=True, shortcut='F4' if self.shortcuts else "",
            statustip='Show/hide the macro panel')

        self.goto_cell = Action(self.parent, "Go to cell",
                                self.parent.workflows.view_goto_cell,
                                icon=Icon.goto_cell,
                                shortcut='Ctrl+g' if self.shortcuts else "",
                                statustip='Select a cell and put it into view')

        self.toggle_spell_checker = \
            Action(self.parent, "Toggle spell checker",
                   self.parent.entry_line.on_toggle_spell_check,
                   icon=Icon.check_spelling,
                   checkable=True,
                   statustip='Turn the spell checker in the entry line on/off')

        self.zoom_in = Action(self.parent, "Zoom in",
                              self.parent.grid.on_zoom_in,
                              icon=Icon.zoom_in,
                              shortcut='Ctrl++' if self.shortcuts else "",
                              statustip='Zoom in the grid')

        self.zoom_out = Action(self.parent, "Zoom out",
                               self.parent.grid.on_zoom_out,
                               icon=Icon.zoom_out,
                               shortcut='Ctrl+-' if self.shortcuts else "",
                               statustip='Zoom out the grid')

        self.zoom_1 = Action(self.parent, "Original size",
                             self.parent.grid.on_zoom_1,
                             icon=Icon.zoom_1,
                             shortcut='Ctrl+0' if self.shortcuts else "",
                             statustip='Show grid on standard zoom level')

        self.refresh_cells = \
            Action(self.parent, "Refresh selected cells",
                   self.parent.grid.refresh_selected_frozen_cells,
                   icon=Icon.refresh,
                   shortcut=QKeySequence.Refresh if self.shortcuts else "",
                   statustip='Refresh selected cells even when frozen')

        self.toggle_periodic_updates = \
            Action(self.parent, "Toggle periodic updates",
                   self.parent.on_toggle_refresh_timer,
                   icon=Icon.toggle_periodic_updates, checkable=True,
                   statustip='Toggles periodic updates for frozen cells')

        self.show_frozen = Action(self.parent, "Show frozen",
                                  self.parent.grid.on_show_frozen_pressed,
                                  icon=Icon.show_frozen,
                                  checkable=True,
                                  statustip='Indicates frozen cells with a '
                                            'background crosshatch')

    def create_format_actions(self):
        """actions for Format menu"""

        self.copy_format = Action(self.parent, "&Copy format",
                                  self.parent.workflows.format_copy_format,
                                  icon=Icon.copy_format,
                                  statustip='Copy format of selection to '
                                            'the clipboard')

        self.paste_format = \
            Action(self.parent, "&Paste format",
                   self.parent.workflows.format_paste_format,
                   icon=Icon.paste_format,
                   statustip='Apply format from the clipboard to the selected '
                             'cells')

        self.font = Action(self.parent, "&Font...",
                           self.parent.grid. on_font_dialog,
                           icon=Icon.font_dialog,
                           statustip='Lauch font dialog')

        self.bold = Action(self.parent, "&Bold",
                           self.parent.grid.on_bold_pressed,
                           icon=Icon.bold,
                           shortcut='Ctrl+b' if self.shortcuts else "",
                           checkable=True,
                           statustip='Toggle bold font weight for the '
                                     'selected cells')

        self.italics = Action(self.parent, "&Italics",
                              self.parent.grid.on_italics_pressed,
                              icon=Icon.italics,
                              shortcut='Ctrl+i' if self.shortcuts else "",
                              checkable=True,
                              statustip='Toggle italics font style for the '
                                        'selected cells')

        self.underline = Action(self.parent, "&Underline",
                                self.parent.grid.on_underline_pressed,
                                icon=Icon.underline,
                                shortcut='Ctrl+u' if self.shortcuts else "",
                                checkable=True,
                                statustip='Toggle underline for the '
                                          'selected cells')

        self.strikethrough = Action(self.parent, "&Strikethrough",
                                    self.parent.grid.on_strikethrough_pressed,
                                    icon=Icon.strikethrough,
                                    checkable=True,
                                    statustip='Toggle strikethrough for the '
                                              'selected cells')

        self.text = Action(self.parent, "Text renderer",
                           self.parent.grid.on_text_renderer_pressed,
                           icon=Icon.text,
                           checkable=True,
                           statustip='Show cell results as text (default). '
                                     'Formats affect the whole cell')

        self.markup = Action(self.parent, "Markup renderer",
                             self.parent.grid.on_markup_renderer_pressed,
                             icon=Icon.markup,
                             checkable=True,
                             statustip='Show cell results as markup, which '
                                       'allows partly formatted output')

        self.image = Action(self.parent, "Image renderer",
                            self.parent.grid.on_image_renderer_pressed,
                            icon=Icon.image,
                            checkable=True,
                            statustip='Show cell results as image. A numpy '
                                      'array of shape (x, y, 3) '
                                      'is expected')
        if matplotlib_figure is not None:
            self.matplotlib = \
                Action(self.parent, "Matplotlib chart renderer",
                       self.parent.grid.on_matplotlib_renderer_pressed,
                       icon=Icon.matplotlib,
                       checkable=True,
                       statustip='Show cell results as matplotlib chart. A '
                                 'numpy array of shape (x, y, 3) is expected')

        renderer_group = QActionGroup(self.parent)
        renderer_group.addAction(self.text)
        renderer_group.addAction(self.markup)
        renderer_group.addAction(self.image)
        if matplotlib_figure is not None:
            renderer_group.addAction(self.matplotlib)

        self.text_color = Action(
            self.parent, "Text color...",
            self.parent.widgets.text_color_button.on_pressed,
            icon=Icon.text_color, statustip='Lauch text color dialog')

        self.line_color = Action(
            self.parent, "Line color...",
            self.parent.widgets.line_color_button.on_pressed,
            icon=Icon.line_color, statustip='Lauch line color dialog')

        self.background_color = Action(
            self.parent, "Background color...",
            self.parent.widgets.background_color_button.on_pressed,
            icon=Icon.background_color,
            statustip='Lauch background color dialog')

        self.freeze_cell = Action(self.parent, "Freeze cell",
                                  self.parent.grid.on_freeze_pressed,
                                  icon=Icon.freeze,
                                  checkable=True,
                                  statustip='Freeze the selected cell so that '
                                            'is is only updated when <F5> is '
                                            'pressed')

        self.lock_cell = Action(self.parent, "Lock cell",
                                self.parent.grid.on_lock_pressed,
                                icon=Icon.lock,
                                checkable=True,
                                statustip='Lock cell so that its code '
                                          'cannot be changed')

        self.button_cell = Action(self.parent, "Button cell",
                                  self.parent.grid.on_button_cell_pressed,
                                  icon=Icon.button,
                                  checkable=True,
                                  statustip='Make cell a button cell that is '
                                            'executed only when pressed')

        self.merge_cells = Action(self.parent, "Merge cells",
                                  self.parent.grid.on_merge_pressed,
                                  icon=Icon.merge_cells,
                                  checkable=True,
                                  statustip='Merge/unmerge selected cells')

        self.rotate_0 = Action(self.parent, "0°",
                               self.parent.grid.on_rotate_0,
                               icon=Icon.rotate_0,
                               checkable=True,
                               statustip='Set text rotation to 0°')

        self.rotate_90 = Action(self.parent, "90°",
                                self.parent.grid.on_rotate_90,
                                icon=Icon.rotate_90,
                                checkable=True,
                                statustip='Set text rotation to 90°')

        self.rotate_180 = Action(self.parent, "180°",
                                 self.parent.grid.on_rotate_180,
                                 icon=Icon.rotate_180,
                                 checkable=True,
                                 statustip='Set text rotation to 180°')

        self.rotate_270 = Action(self.parent, "270°",
                                 self.parent.grid.on_rotate_270,
                                 icon=Icon.rotate_270,
                                 checkable=True,
                                 statustip='Set text rotation to 270°')

        rotate_group = QActionGroup(self.parent)
        rotate_group.addAction(self.rotate_0)
        rotate_group.addAction(self.rotate_90)
        rotate_group.addAction(self.rotate_180)
        rotate_group.addAction(self.rotate_270)

        self.justify_left = Action(self.parent, "Left",
                                   self.parent.grid.on_justify_left,
                                   icon=Icon.justify_left,
                                   checkable=True,
                                   statustip='Display cell result text '
                                             'left justified')

        self.justify_center = Action(self.parent, "Center",
                                     self.parent.grid.on_justify_center,
                                     checkable=True,
                                     icon=Icon.justify_center,
                                     statustip='Display cell result text '
                                               'centered')

        self.justify_right = Action(self.parent, "Right",
                                    self.parent.grid.on_justify_right,
                                    checkable=True,
                                    icon=Icon.justify_right,
                                    statustip='Display cell result text '
                                              'right justified')

        self.justify_fill = Action(self.parent, "Fill",
                                   self.parent.grid.on_justify_fill,
                                   icon=Icon.justify_fill,
                                   checkable=True,
                                   statustip='Display cell result text '
                                             'filled into the cell')

        justify_group = QActionGroup(self.parent)
        justify_group.addAction(self.justify_left)
        justify_group.addAction(self.justify_center)
        justify_group.addAction(self.justify_right)
        justify_group.addAction(self.justify_fill)

        self.align_top = Action(self.parent, "Top",
                                self.parent.grid.on_align_top,
                                icon=Icon.align_top,
                                checkable=True,
                                statustip='Align cell result at the top of '
                                          'the cell')

        self.align_center = Action(self.parent, "Center",
                                   self.parent.grid.on_align_middle,
                                   icon=Icon.align_center,
                                   checkable=True,
                                   statustip='Center cell result within '
                                             'the cell')

        self.align_bottom = Action(self.parent, "Bottom",
                                   self.parent.grid.on_align_bottom,
                                   icon=Icon.align_bottom,
                                   checkable=True,
                                   statustip='Align cell result at the '
                                             'bottom of the cell')

        align_group = QActionGroup(self.parent)
        align_group.addAction(self.align_top)
        align_group.addAction(self.align_center)
        align_group.addAction(self.align_bottom)

        self.format_borders_all = \
            Action(self.parent, "All borders",
                   self.parent.grid.on_border_choice,
                   icon=Icon.format_borders_all, checkable=True,
                   statustip='Format all borders of selection')

        self.format_borders_top = \
            Action(self.parent, "Top border",
                   self.parent.grid.on_border_choice,
                   icon=Icon.format_borders_top, checkable=True,
                   statustip='Format top border of selection')

        self.format_borders_bottom = \
            Action(self.parent, "Bottom border",
                   self.parent.grid.on_border_choice,
                   icon=Icon.format_borders_bottom, checkable=True,
                   statustip='Format bottom border of selection')

        self.format_borders_left = \
            Action(self.parent, "Left border",
                   self.parent.grid.on_border_choice,
                   icon=Icon.format_borders_left, checkable=True,
                   statustip='Format left border of selection')

        self.format_borders_right = \
            Action(self.parent, "Right border",
                   self.parent.grid.on_border_choice,
                   icon=Icon.format_borders_right, checkable=True,
                   statustip='Format right border of selection')

        self.format_borders_outer = \
            Action(self.parent, "Outer borders",
                   self.parent.grid.on_border_choice,
                   icon=Icon.format_borders_outer, checkable=True,
                   statustip='Format outer borders of selection')

        self.format_borders_inner = \
            Action(self.parent, "Inner borders",
                   self.parent.grid.on_border_choice,
                   icon=Icon.format_borders_inner, checkable=True,
                   statustip='Format inner borders of selection')

        self.format_borders_top_bottom = \
            Action(self.parent, "Top and bottom borders",
                   self.parent.grid.on_border_choice,
                   icon=Icon.format_borders_top_bottom, checkable=True,
                   statustip='Format top and bottom borders of selection')

        self.border_group = QActionGroup(self.parent)
        self.border_group.addAction(self.format_borders_all)
        self.border_group.addAction(self.format_borders_top)
        self.border_group.addAction(self.format_borders_bottom)
        self.border_group.addAction(self.format_borders_left)
        self.border_group.addAction(self.format_borders_right)
        self.border_group.addAction(self.format_borders_outer)
        self.border_group.addAction(self.format_borders_inner)
        self.border_group.addAction(self.format_borders_top_bottom)
        self.format_borders_all.setChecked(True)

        self.format_borders_0 = Action(self.parent, "Border width 0",
                                       self.parent.grid.on_borderwidth,
                                       icon=Icon.format_borders_0,
                                       statustip='Set border width to 0')

        self.format_borders_1 = Action(self.parent, "Border width 1",
                                       self.parent.grid.on_borderwidth,
                                       icon=Icon.format_borders_1,
                                       statustip='Set border width to 1')

        self.format_borders_2 = Action(self.parent, "Border width 2",
                                       self.parent.grid.on_borderwidth,
                                       icon=Icon.format_borders_2,
                                       statustip='Set border width to 2')

        self.format_borders_4 = Action(self.parent, "Border width 3",
                                       self.parent.grid.on_borderwidth,
                                       icon=Icon.format_borders_4,
                                       statustip='Set border width to 3')

        self.format_borders_8 = Action(self.parent, "Border width 4",
                                       self.parent.grid.on_borderwidth,
                                       icon=Icon.format_borders_8,
                                       statustip='Set border width to 4')

        self.format_borders_16 = Action(self.parent, "Border width 6",
                                        self.parent.grid.on_borderwidth,
                                        icon=Icon.format_borders_16,
                                        statustip='Set border width to 6')

        self.format_borders_32 = Action(self.parent, "Border width 8",
                                        self.parent.grid.on_borderwidth,
                                        icon=Icon.format_borders_32,
                                        statustip='Set border width to 8')

        self.format_borders_64 = Action(self.parent, "Border width 12",
                                        self.parent.grid.on_borderwidth,
                                        icon=Icon.format_borders_64,
                                        statustip='Set border width to 12')

        self.border_width_group = QActionGroup(self.parent)
        self.border_width_group.addAction(self.format_borders_0)
        self.border_width_group.addAction(self.format_borders_1)
        self.border_width_group.addAction(self.format_borders_2)
        self.border_width_group.addAction(self.format_borders_4)
        self.border_width_group.addAction(self.format_borders_8)
        self.border_width_group.addAction(self.format_borders_16)
        self.border_width_group.addAction(self.format_borders_32)
        self.border_width_group.addAction(self.format_borders_64)
        self.format_borders_1.setChecked(True)

    def create_macro_actions(self):
        """Create actions for Macro menu"""

        self.insert_image = Action(self.parent, "Insert image...",
                                   self.parent.workflows.macro_insert_image,
                                   icon=Icon.insert_image,
                                   statustip='Load an image from a file '
                                             'into a cell')

        self.insert_chart = Action(self.parent, "Insert chart...",
                                   self.parent.workflows.macro_insert_chart,
                                   icon=Icon.insert_chart,
                                   statustip='Create and display matplotlib '
                                             'chart')

        self.insert_sum = Action(self.parent, "Insert sum",
                                 self.parent.workflows.macro_insert_sum,
                                 icon=Icon.insert_sum,
                                 statustip='Insert sum of selection into the'
                                           'cell below the bottom right cell '
                                           'of the selection')

    def create_help_actions(self):
        """actions for Help menu"""

        self.manual = Action(self.parent, "Manual...",
                             self.parent.on_manual,
                             icon=Icon.help,
                             shortcut='F1' if self.shortcuts else "",
                             statustip='Display the pyspread manual')

        self.tutorial = Action(self.parent, "Tutorial...",
                               self.parent.on_tutorial,
                               icon=Icon.tutorial,
                               statustip='Display a pyspread tutorial')

        self.dependencies = Action(self.parent, "Dependencies...",
                                   self.parent.on_dependencies,
                                   icon=Icon.dependencies,
                                   statustip='List and install dependencies')

        self.about = Action(self.parent, "About pyspread...",
                            self.parent.on_about,
                            icon=Icon.pyspread,
                            statustip='About pyspread',
                            role=QAction.AboutRole)

    def disable_unavailable(self):
        """Disables unavailable menu items e.g. due to missing dependencies"""

        if enchant is None:
            self.toggle_spell_checker.setEnabled(False)


class ChartDialogActions(AttrDict):
    """QActions for chart dialog

    Reads out template files in share/templates/matplotlib and
    adds a QAction for each template.

    """

    def __init__(self, parent: QWidget):
        """
        :param parent: The parent object

        """

        super().__init__()
        self.parent = parent
        self._add_chart_template_actions()

    def _add_chart_template_actions(self):
        """Adds actions for chart dialog toolbar"""

        self.chart_pie_1_1 = Action(self.parent, "Pie chart",
                                    self.parent.on_template,
                                    icon=Icon.chart_pie_1_1,
                                    statustip='Insert code for pie chart')
        self.chart_pie_1_1.setData("chart_pie_1_1.py")

        self.chart_ring_1_1 = Action(self.parent, "Ring chart",
                                     self.parent.on_template,
                                     icon=Icon.chart_ring_1_1,
                                     statustip='Insert code for ring chart')
        self.chart_ring_1_1.setData("chart_ring_1_1.py")

        self.chart_line_1_1 = Action(self.parent, "Line chart",
                                     self.parent.on_template,
                                     icon=Icon.chart_line_1_1,
                                     statustip='Insert code for line chart')
        self.chart_line_1_1.setData("chart_line_1_1.py")

        self.chart_polar_1_1 = Action(self.parent, "Polar chart",
                                      self.parent.on_template,
                                      icon=Icon.chart_polar_1_1,
                                      statustip='Insert code for polar '
                                      'coordinates line chart')
        self.chart_polar_1_1.setData("chart_polar_1_1.py")

        self.chart_area_1_1 = Action(self.parent, "Area chart",
                                     self.parent.on_template,
                                     icon=Icon.chart_area_1_1,
                                     statustip='Insert code for area chart')
        self.chart_area_1_1.setData("chart_area_1_1.py")

        self.chart_column_1_1 = Action(self.parent, "Grouped column chart",
                                       self.parent.on_template,
                                       icon=Icon.chart_column_1_1,
                                       statustip='Insert code for grouped '
                                                 'column chart')
        self.chart_column_1_1.setData("chart_column_1_1.py")

        self.chart_column_1_2 = Action(self.parent, "Stacked column chart",
                                       self.parent.on_template,
                                       icon=Icon.chart_column_1_2,
                                       statustip='Insert code for stacked '
                                                 'column chart')
        self.chart_column_1_2.setData("chart_column_1_2.py")

        self.chart_bar_1_3 = \
            Action(self.parent, "Normalized stacked bar chart",
                   self.parent.on_template, icon=Icon.chart_bar_1_3,
                   statustip='Insert code for normalized stacked bar chart')
        self.chart_bar_1_3.setData("chart_bar_1_3.py")

        self.chart_scatter_1_1 = Action(self.parent, "Scatter chart",
                                        self.parent.on_template,
                                        icon=Icon.chart_scatter_1_1,
                                        statustip='Insert code for a scatter '
                                                  'plot')
        self.chart_scatter_1_1.setData("chart_scatter_1_1.py")

        self.chart_bubble_1_1 = Action(self.parent, "Bubble chart",
                                       self.parent.on_template,
                                       icon=Icon.chart_bubble_1_1,
                                       statustip='Insert code for a bubble '
                                       'plot that is a scatter plot with '
                                       'individual point sizes')
        self.chart_bubble_1_1.setData("chart_bubble_1_1.py")

        self.chart_boxplot_2_2 = Action(self.parent, "Boxplot chart",
                                        self.parent.on_template,
                                        icon=Icon.chart_boxplot_2_2,
                                        statustip='Insert code for boxplot '
                                        'chart')
        self.chart_boxplot_2_2.setData("chart_boxplot_2_2.py")

        self.chart_histogram_1_1 = Action(self.parent, "Histogram chart",
                                          self.parent.on_template,
                                          icon=Icon.chart_histogram_1_1,
                                          statustip='Insert code for '
                                                    'histogram')
        self.chart_histogram_1_1.setData("chart_histogram_1_1.py")

        self.chart_histogram_1_4 = Action(self.parent,
                                          "Multiple histogram charts",
                                          self.parent.on_template,
                                          icon=Icon.chart_histogram_1_4,
                                          statustip='Insert code for multiple '
                                                    'histogram charts')
        self.chart_histogram_1_4.setData("chart_histogram_1_4.py")

        self.chart_scatterhist_1_1 = Action(self.parent,
                                            "Scatter and histogram chart",
                                            self.parent.on_template,
                                            icon=Icon.chart_scatterhist_1_1,
                                            statustip='Insert code for scatter'
                                                      ' plot with histogram'
                                                      ' for each axis')
        self.chart_scatterhist_1_1.setData("chart_scatterhist_1_1.py")

        self.chart_matrix_1_1 = Action(self.parent, "Matrix chart",
                                       self.parent.on_template,
                                       icon=Icon.chart_matrix_1_1,
                                       statustip='Insert code for matrix '
                                                 'chart')
        self.chart_matrix_1_1.setData("chart_matrix_1_1.py")

        self.chart_contour_1_2 = Action(self.parent, "Contour chart",
                                        self.parent.on_template,
                                        icon=Icon.chart_contour_1_2,
                                        statustip='Insert code for contour '
                                                  'chart')
        self.chart_contour_1_2.setData("chart_contour_1_2.py")

        self.chart_surface_2_1 = Action(self.parent, "Surface chart",
                                        self.parent.on_template,
                                        icon=Icon.chart_surface_2_1,
                                        statustip='Insert code for surface '
                                                  'chart')
        self.chart_surface_2_1.setData("chart_surface_2_1.py")

        self.chart_plotnine_geom_bar_1_1 = \
            Action(self.parent, "Plotnine geom_bar chart",
                   self.parent.on_template,
                   icon=Icon.chart_plotnine_geom_bar_1_1,
                   statustip='Insert code for plotnine geom_bar chart')
        self.chart_plotnine_geom_bar_1_1.setData(
            "chart_plotnine_geom_bar_1_1.py")

        self.chart_r_graphics_barplot_1_1 = \
            Action(self.parent, "R graphics barplot chart",
                   self.parent.on_template,
                   icon=Icon.chart_r_graphics_barplot_1_1,
                   statustip='Insert code for R graphics barplot chart')
        self.chart_r_graphics_barplot_1_1.setData(
            "chart_r_graphics_barplot_1_1.py")

        self.chart_r_ggplot2_geom_boxplot_1_2 = \
            Action(self.parent, "R ggplot2 geom_boxplot chart",
                   self.parent.on_template,
                   icon=Icon.chart_r_ggplot2_geom_boxplot_1_2,
                   statustip='Insert code for R ggplot2 geom_boxplot chart')
        self.chart_r_ggplot2_geom_boxplot_1_2.setData(
            "chart_r_ggplot2_geom_boxplot_1_2.py")

        self.chart_r_ggplot2_geom_point_1_1 = \
            Action(self.parent, "R ggplot2 geom_point chart",
                   self.parent.on_template,
                   icon=Icon.chart_r_ggplot2_geom_point_1_1,
                   statustip='Insert code for R ggplot2 geom_point chart')
        self.chart_r_ggplot2_geom_point_1_1.setData(
            "chart_r_ggplot2_geom_point_1_1.py")

        self.chart_r_lattice_xyplot_1_1 = \
            Action(self.parent, "R lattice xyplot chart",
                   self.parent.on_template,
                   icon=Icon.chart_r_lattice_xyplot_1_1,
                   statustip='Insert code for R lattice xyplot chart')
        self.chart_r_lattice_xyplot_1_1.setData(
            "chart_r_lattice_xyplot_1_1.py")

        self.chart_r_ggplot2_geom_density2d_1_2 = \
            Action(self.parent, "R ggplot2 geom_density2d chart",
                   self.parent.on_template,
                   icon=Icon.chart_r_ggplot2_geom_density2d_1_2,
                   statustip='Insert code for R ggplot2 geom_density2d chart')
        self.chart_r_ggplot2_geom_density2d_1_2.setData(
            "chart_r_ggplot2_geom_density2d_1_2.py")

        self.chart_r_lattice_wireframe_2_1 = \
            Action(self.parent, "R lattice wireframe chart",
                   self.parent.on_template,
                   icon=Icon.chart_r_lattice_wireframe_2_1,
                   statustip='Insert code for surface chart')
        self.chart_r_lattice_wireframe_2_1.setData(
            "chart_r_lattice_wireframe_2_1.py")


class SpellTextEditActions(AttrDict):
    """Holds QActions for SpellTextEdit"""

    def __init__(self, parent: QWidget):
        """
        :param parent: The parent object, normally :class:`pyspread.MainWindow`

        """

        super().__init__()
        self.parent = parent

        self.toggle_line_numbers = Action(self.parent, "Line numbers",
                                          self.parent.show_line_numbers,
                                          checkable=True,
                                          statustip='Show/hide line numbers')



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------


"""

**Unit tests for file_helpers.py**

"""

import pytest
from ..file_helpers import linecount


param_test_linecount = [
    (b"", 0),
    (b"\n", 1),
    (b"Test1\nTest2", 1),
    (b"Test1\nTest2\n", 2),
    (b" \n" * 1000, 1000),
    (b"\n" * 2**20, 2**20),
]


@pytest.mark.parametrize("content, res", param_test_linecount)
def test_linecount(content, res, tmp_path):
    """Unit test for linecount"""

    testfile_path = tmp_path / "test"
    testfile_path.write_bytes(content)
    with open(testfile_path, "rb") as testfile:
        assert linecount(testfile) == res



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

Grid selection representation


**Provides**

* :class:`Selection`: Represents grid selection independently from PyQt

"""

from builtins import zip, range, object
from typing import Generator, List, Tuple


class Selection:
    """Represents grid selection"""

    def __init__(self,
                 block_top_left: List[Tuple[int, int]],
                 block_bottom_right: List[Tuple[int, int]],
                 rows: List[int],
                 columns: List[int],
                 cells: List[Tuple[int, int]]):
        """
        :param block_top_left: Top left edges of all selection rectangles
        :param block_bottom_right: Top left edges of all selection rectangles
        :param rows: Selected rows
        :param columns: Selected columns
        :param cells: Individually selected cells as list of (row, column)

        """

        self.block_tl = block_top_left
        self.block_br = block_bottom_right
        self.rows = rows
        self.columns = columns
        self.cells = cells

    def __bool__(self) -> bool:
        """
        :return: True iif any attribute is non-empty

        """

        return any(self.parameters)

    def __repr__(self) -> str:
        """
        :return: String output for printing selection

        """

        return f"Selection{self.parameters}"

    def __eq__(self, other):
        """Eqality check

        Selections are equal iif the order of each attribute is equal
        because order precedence may change the selection outcome in the grid.

        :param other: Other selection for equality comparison
        :type other: Selection
        :return: True if self and other selection are equal

        """

        attrs = "block_tl", "block_br", "rows", "columns", "cells"

        return all(getattr(self, at) == getattr(other, at) for at in attrs)

    def __contains__(self, cell: Tuple[int, int]):
        """Check if cell is included in self

        :param cell: Index of cell to be checked
        :return: True iif cell is in selection

        """

        if len(cell) != 2:
            raise Warning("Key length is not 2. Returning None.")
            return

        cell_row, cell_col = cell

        # Block selections
        for top_left, bottom_right in zip(self.block_tl, self.block_br):
            top, left = top_left
            bottom, right = bottom_right

            if top is None:
                top = 0

            if left is None:
                left = 0

            if bottom is None:
                bottom = cell_row

            if right is None:
                right = cell_col

            if top <= cell_row <= bottom and left <= cell_col <= right:
                return True

        # Row and column selections

        if cell_row in self.rows or cell_col in self.columns:
            return True

        # Cell selections
        if cell in self.cells:
            return True

        return False

    def __add__(self, value: Tuple[int, int]):
        """Shifts selection down and / or right

        :param value: Number of rows / columns to be shifted down / right
        :return: Shifted selection
        :rtype: Selection

        """

        def shifted_block(block0: int, block1: int, delta_row: int,
                          delta_col: int) -> Tuple[int, int]:
            """Returns shifted block"""

            try:
                row = block0 + delta_row
            except TypeError:
                row = block0

            try:
                col = block1 + delta_col
            except TypeError:
                col = block1

            return row, col

        delta_row, delta_col = value

        block_tl = [shifted_block(top, left, delta_row, delta_col)
                    for top, left in self.block_tl]

        block_br = [shifted_block(bottom, right, delta_row, delta_col)
                    for bottom, right in self.block_br]

        rows = [row + delta_row for row in self.rows]
        columns = [col + delta_col for col in self.columns]
        cells = [(r + delta_row, c + delta_col) for r, c in self.cells]

        return Selection(block_tl, block_br, rows, columns, cells)

    def __and__(self, other):
        """Returns intersection selection of self and other

        :param other: Other selection for intersecting
        :type other: Selection
        :return: Intersection selection
        :rtype: Selection

        """

        block_tl = []
        block_br = []
        rows = []
        columns = []
        cells = []

        # Blocks
        # Check cells in block: If all are in other, add block else add cells

        for block in zip(self.block_tl, self.block_br):
            if block[0] in other.block_tl and block[1] in other.block_br:
                block_tl.append(block[0])
                block_br.append(block[1])
            else:
                block_cells = []
                for row in range(block[0][0], block[1][0] + 1):
                    for col in range(block[0][1], block[1][1] + 1):
                        cell = row, col
                        if cell in other:
                            block_cells.append(cell)

                if len(block_cells) == (block[1][0] + 1 - block[0][0]) * \
                                       (block[1][1] + 1 - block[0][1]):
                    block_tl.append(block[0])
                    block_br.append(block[1])
                else:
                    cells.extend(block_cells)

        # Rows
        # If a row/col is selected in self and other then add it.
        # Otherwise, add all cells in the respective row/col that are in other.

        for row in self.rows:
            if row in other.rows:
                rows.append(row)
            else:
                for block in zip(other.block_tl, other.block_br):
                    if block[0][0] <= row <= block[1][0]:
                        block_tl.append((row, block[0][1]))
                        block_br.append((row, block[1][1]))

                for cell in other.cells:
                    if cell[0] == row and cell not in cells:
                        cells.append(cell)

        # Columns

        for col in self.columns:
            if col in other.columns:
                columns.append(col)
            else:
                for block in zip(other.block_tl, other.block_br):
                    if block[0][1] <= col <= block[1][1]:
                        block_tl.append((block[0][0], col))
                        block_br.append((block[1][0], col))

                for cell in other.cells:
                    if cell[1] == col and cell not in cells:
                        cells.append(cell)

        # Cells

        for cell in self.cells:
            if cell in other and cell not in cells:
                cells.append(cell)

        cells = list(set(cells))

        return Selection(block_tl, block_br, rows, columns, cells)

    # Parameter access

    @property
    def parameters(self) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]],
                                  List[int], List[int], List[Tuple[int, int]]]:
        """

        :return: Tuple of selection parameters of self
                 (self.block_tl, self.block_br, self.rows, self.columns,
                  self.cells)

        """

        return (self.block_tl, self.block_br, self.rows, self.columns,
                self.cells)

    def insert(self, point: int, number: int, axis: int):
        """Inserts number of rows/columns/tables into selection

        Insertion takes place at point on axis.

        :param point: At this point the rows/columns are inserted or deleted
        :param number: Number of rows/columns to be inserted
                       or deleted if negative
        :param axis: If 0, rows are affected, if 1, columns, axis in 0, 1

        """

        def build_tuple_list(source_list, point, number, axis):
            """Returns adjusted tuple list for single cells"""

            target_list = []

            for tl in source_list:
                tl_list = list(tl)
                if tl[axis] >= point:
                    tl_list[axis] += number
                target_list.append(tuple(tl_list))

            return target_list

        if number == 0:
            return

        self.block_tl = build_tuple_list(self.block_tl, point, number, axis)

        self.block_br = build_tuple_list(self.block_br, point, number, axis)

        if axis == 0:
            self.rows = [row + number if row >= point else row
                         for row in self.rows]
        elif axis == 1:
            self.columns = [column + number if column >= point else column
                            for column in self.columns]
        else:
            raise ValueError("Axis not in [0, 1]")

        self.cells = build_tuple_list(self.cells, point, number, axis)

    def get_bbox(self) -> Tuple[Tuple[int, int], Tuple[int, int]]:
        """Returns bounding box

        A bounding box is the smallest rectangle that contains all selections.
        Non-specified boundaries are None.

        :return: ((top, left), (bottom, right)) of bounding box

        """

        bb_top, bb_left, bb_bottom, bb_right = [None] * 4

        # Block selections

        for top_left, bottom_right in zip(self.block_tl, self.block_br):
            top, left = top_left
            bottom, right = bottom_right

            if bb_top is None or bb_top > top:
                bb_top = top
            if bb_left is None or bb_left > left:
                bb_left = left
            if bb_bottom is None or bb_bottom < bottom:
                bb_bottom = bottom
            if bb_right is None or bb_right < right:
                bb_right = right

        # Row and column selections

        for row in self.rows:
            if bb_top is None or bb_top > row:
                bb_top = row
            if bb_bottom is None or bb_bottom < row:
                bb_bottom = row

        for col in self.columns:
            if bb_left is None or bb_left > col:
                bb_left = col
            if bb_right is None or bb_right < col:
                bb_right = col

        # Cell selections

        for cell in self.cells:
            cell_row, cell_col = cell

            if bb_top is None or bb_top > cell_row:
                bb_top = cell_row
            if bb_left is None or bb_left > cell_col:
                bb_left = cell_col
            if bb_bottom is None or bb_bottom < cell_row:
                bb_bottom = cell_row
            if bb_right is None or bb_right < cell_col:
                bb_right = cell_col

        if self.rows:
            bb_left = bb_right = None

        if self.columns:
            bb_top = bb_bottom = None

        return ((bb_top, bb_left), (bb_bottom, bb_right))

    def get_grid_bbox(self, shape: Tuple[int, int, int]
                      ) -> Tuple[Tuple[int, int], Tuple[int, int]]:
        """Returns bounding box within grid shape limits

        A bounding box is the smallest rectangle that contains all selections.
        Non-specified boundaries are filled i from size.

        :param shape: Grid shape
        :return: ((top, left), (bottom, right)) of bounding box

        """

        (bb_top, bb_left), (bb_bottom, bb_right) = self.get_bbox()

        if bb_top is None:
            bb_top = 0
        if bb_left is None:
            bb_left = 0
        if bb_bottom is None:
            bb_bottom = shape[0]
        if bb_right is None:
            bb_right = shape[1]

        return ((bb_top, bb_left), (bb_bottom, bb_right))

    def get_absolute_access_string(self, shape: Tuple[int, int, int],
                                   table: int) -> str:
        """Get access string for absolute addressing

        :param shape: Grid shape, for which the generated keys are valid
        :param table: Table for all returned keys. Must be valid table in shape
        :return: String, with which the selection can be accessed

        """

        rows, columns, _ = shape
        strings = []

        # Block selections
        for (top, left), (bottom, right) in zip(self.block_tl, self.block_br):
            strings += [f"[(r, c, {table})"
                        f" for r in range({top}, {bottom + 1})"
                        f" for c in range({left}, {right + 1})]"]

        # Fully selected rows
        for row in self.rows:
            strings += [f"[({row}, c, {table}) for c in range({columns})]"]

        # Fully selected columns
        for column in self.columns:
            strings += [f"[(r, {column}, {table}) for r in range({rows})]"]

        # Single cells
        for row, column in self.cells:
            strings += [f"[({row}, {column}, {table})]"]

        if not strings:
            return ""

        if len(self.cells) == 1 and len(strings) == 1:
            return f"S[{strings[0][2:-2]}]"

        key_string = " + ".join(strings)
        return f"[S[key] for key in {key_string} if S[key] is not None]"

    def get_relative_access_string(self, shape: Tuple[int, int, int],
                                   current: Tuple[int, int, int]) -> str:
        """Get access string relative to current cell

        It is assumed that the selected cells are in the same table as the
        current cell.

        :param shape: Grid shape, for which the generated keys are valid
        :param current: Current cell for relative addressing
        :return: String, with which the selection can be accessed

        """

        rows, columns, tables = shape
        crow, ccolumn, ctable = current

        strings = []

        # Block selections
        for (top, left), (bottom, right) in zip(self.block_tl, self.block_br):
            strings += [
                "[(X + dr, Y + dc, Z)" +
                f" for dr in range({top - crow}, {bottom - crow + 1})"
                f" for dc in range({left - ccolumn}, {right - ccolumn + 1})]"]

        # Fully selected rows
        for row in self.rows:
            strings += [
                f"[(X + {row - crow}, c, Z) for c in range({columns})]"]

        # Fully selected columns
        for column in self.columns:
            strings += [f"[(r, {column - ccolumn}, Z) for r in range({rows})]"]

        # Single cells
        for row, column in self.cells:
            strings += [f"[(X + {row-crow}, Y + {column-ccolumn}, Z)]"]

        key_string = " + ".join(strings)

        if not strings:
            return ""

        if len(self.cells) == 1 and len(strings) == 1:
            return f"S[{strings[0][2:-2]}]"

        return f"[S[key] for key in {key_string} if S[key] is not None]"

    def shifted(self, rows: int, columns: int):
        """Get a shifted selection

        Negative values for rows and columns may result in a selection
        that addresses negative cells.

        :param rows: Number of rows that the selection is shifted down
        :param columns: Number of columns that the selection is shifted right
        :return: New selection that is shifted by rows and columns
        :rtype: Selection

        """

        shifted_block_tl = [(row + rows, col + columns)
                            for row, col in self.block_tl]
        shifted_block_br = [(row + rows, col + columns)
                            for row, col in self.block_br]
        shifted_rows = [row + rows for row in self.rows]
        shifted_columns = [col + columns for col in self.columns]
        shifted_cells = [(row + rows, col + columns)
                         for row, col in self.cells]

        return Selection(shifted_block_tl, shifted_block_br, shifted_rows,
                         shifted_columns, shifted_cells)

    def get_right_borders_selection(self, border_choice: str,
                                    shape: Tuple[int, int, int]):
        """Get selection of cells, for which the right border attributes
        need to be adjusted on border line and border color changes.

        border_choice names are:
         * "All borders"
         * "Top border"
         * "Bottom border"
         * "Left border"
         * "Right border"
         * "Outer borders"
         * "Inner borders"
         * "Top and bottom borders"

        :param border_choice: Border choice name
        :return: Selection of cells that need to be adjusted on border change
        :rtype: Selection

        """

        (top, left), (bottom, right) = self.get_grid_bbox(shape)

        if border_choice == "All borders":
            return Selection([(top, left-1)], [(bottom, right)], [], [], [])

        if border_choice in ("Top border", "Bottom border",
                             "Top and bottom borders"):
            return Selection([], [], [], [], [])

        if border_choice == "Left border":
            return Selection([(top, left-1)], [(bottom, left-1)], [], [], [])

        if border_choice == "Right border":
            return Selection([(top, right)], [(bottom, right)], [], [], [])

        if border_choice == "Outer borders":
            return Selection([(top, right), (top, left-1)],
                             [(bottom, right), (bottom, left-1)], [], [], [])

        if border_choice == "Inner borders":
            return Selection([(top, left)], [(bottom, right-1)], [], [], [])

        raise ValueError(f"border_choice {border_choice} unknown.")

    def get_bottom_borders_selection(self, border_choice: str,
                                     shape: Tuple[int, int, int]):
        """Get selection of cells, for which the bottom border attributes
        need to be adjusted on border line and border color changes.

        border_choice names are:
         * "All borders"
         * "Top border"
         * "Bottom border"
         * "Left border"
         * "Right border"
         * "Outer borders"
         * "Inner borders"
         * "Top and bottom borders"

        :param border_choice: Border choice name
        :return: Selection of cells that need to be adjusted on border change
        :rtype: Selection

        """

        (top, left), (bottom, right) = self.get_grid_bbox(shape)

        if border_choice == "All borders":
            return Selection([(top-1, left)], [(bottom, right)], [], [], [])

        if border_choice == "Top border":
            return Selection([(top-1, left)], [(top-1, right)], [], [], [])

        if border_choice == "Bottom border":
            return Selection([(bottom, left)], [(bottom, right)], [], [], [])

        if border_choice in ("Left border", "Right border"):
            return Selection([], [], [], [], [])

        if border_choice == "Outer borders":
            return Selection([(top-1, left), (bottom, left)],
                             [(top-1, right), (bottom, right)], [], [], [])

        if border_choice == "Inner borders":
            return Selection([(top, left)], [(bottom-1, right)], [], [], [])

        if border_choice == "Top and bottom borders":
            return Selection([(top-1, left), (bottom, left)],
                             [(top-1, right), (bottom, right)], [], [], [])

        raise ValueError(f"border_choice {border_choice} unknown.")

    def single_cell_selected(self) -> bool:
        """
        :return: True iif a single cell is selected via self.cells

        """

        return len(self.cells) == 1 and not any((self.block_tl, self.block_br,
                                                 self.rows, self.columns))

    def cell_generator(self, shape, table=None) -> Generator:
        """Returns a generator of cell key tuples

        :param shape: Grid shape
        :param table: Third component of each returned key

        If table is None 2-tuples (row, column) are yielded else 3-tuples

        """

        rows, columns, tables = shape

        (top, left), (bottom, right) = self.get_grid_bbox(shape)
        bottom = min(bottom, rows - 1)
        right = min(right, columns - 1)

        for row in range(top, bottom + 1):
            for column in range(left, right + 1):
                if (row, column) in self:
                    if table is None:
                        yield row, column
                    elif table < tables:
                        yield row, column, table



fig = Figure()
ax = fig.add_axes([.2,.2, .7, .7], projection='polar')

r = numpy.arange(0, 2, 0.01)
theta = 2 * numpy.pi * r

ax.plot(theta, r)
ax.set_rmax(2)
ax.set_rticks([0.5, 1, 1.5, 2])  # Less radial ticks
ax.set_rlabel_position(-22.5)  # Move radial labels away from plotted line
ax.grid(True)

ax.set(aspect="equal")
ax.set_title("Line plot on polar axis", pad=20)

fig



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

 * :func:`quote`
 * :func:`wrap_text`

"""

import textwrap


def quote(code: str) -> str:
    """Quote code

    :param code: Code to be quoted
    :return: Quoted code if not already quoted and quoting possible

    """

    starts_and_ends = [
        ("'", "'"),
        ('"', '"'),
        ("u'", "'"),
        ('u"', '"'),
        ("b'", "'"),
        ('b"', '"'),
        ("r'", "'"),
        ('r"', '"'),
    ]

    if code is None or not (isinstance(code, bytes) or isinstance(code, str)):
        return code

    code = code.strip()

    if code and not (code[0],  code[-1]) in starts_and_ends:
        return repr(code)
    else:
        return code


def wrap_text(text, width=80, maxlen=2000):
    """Wrap text to line width

    :param text: The text to be wrapped
    :param width: Width of the text to be wrapped
    :param maxlen: Maximum total text length before text in truncated and
                   extended by [...]. If None then truncation is disabled.
    :return: Wrapped text

    """

    if text is None:
        return

    if maxlen is not None and len(text) > maxlen:
        text = text[:maxlen] + "..."
    return "\n".join(textwrap.wrap(text, width=width))



fig = Figure()
ax = fig.add_axes([.2,.2, .7, .7])

# Pie chart, where the slices will be ordered and plotted counter-clockwise:
labels = 'Frogs', 'Hogs', 'Dogs', 'Logs'
sizes = [15, 30, 45, 10]
explode = (0, 0.1, 0, 0)  # only "explode" the 2nd slice (i.e. 'Hogs')

ax.pie(sizes, explode=explode, labels=labels, autopct='%1.1f%%', startangle=90)
ax.axis('equal')  # Equal aspect ratio ensures that pie is drawn as a circle.

ax.set(title='Pie chart')

fig



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

Functions for checking types and type likeness

**Provides**

 * :func:`is_stringlike`
 * :func:`is_svg`
 * :func:`check_shape_validity`

"""

from io import BytesIO
import xml.etree.ElementTree as ET
from typing import Tuple


def is_stringlike(obj: object) -> bool:
    """Is `obj` string like

    :param obj: Object to be checked
    :return: True if obj is instance of str, bytes or bytearray

    """

    return isinstance(obj, (str, bytes, bytearray))


def is_svg(svg_bytes: bytes) -> bool:
    """Checks if svg_bytes is an svg image

    :param svg_bytes: Data to be checked for being an SVG image
    :return: True if svg_bytes seems to be an  SVG image

    """

    tag = None

    with BytesIO(svg_bytes) as svg:
        try:
            for event, el in ET.iterparse(svg, ('start',)):
                tag = el.tag
                break
        except ET.ParseError:
            pass

    return tag == '{http://www.w3.org/2000/svg}svg'


def check_shape_validity(shape: Tuple[int, int, int],
                         maxshape: Tuple[int, int, int]) -> bool:
    """Checks if shape is valid

    :param shape: shape for grid to be checked
    :param maxshape: maximum shape for grid
    :return: True if yes, raises a ValueError otherwise

    """

    try:
        iter(shape)
    except TypeError:
        # not iterable
        raise ValueError("{} is not iterable".format(shape))

    try:
        if len(shape) != 3:
            raise ValueError("len({}) != 3".format(shape))
    except TypeError:
        # no length
        raise ValueError("{} has no length".format(shape))

    if any(ax == 0 for ax in shape):
        raise ValueError("Elements {} equals 0".format(shape))

    if any(ax > axmax for axmax, ax in zip(maxshape, shape)):
        raise ValueError("Grid shape {} exceeds {}.".format(shape, maxshape))

    return True



# -*- coding: utf-8 -*-

# Copyright Martin Manns
# Distributed under the terms of the GNU General Public License

# --------------------------------------------------------------------
# pyspread is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyspread is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyspread.  If not, see <http://www.gnu.org/licenses/>.
# --------------------------------------------------------------------

"""

**Provides**

* :class:`MenuBar`: QMenuBar, the main menubar
* :class:`FileMenu`: File menu for the main menubar
* :class:`EditMenu`: Edit menu for the main menubar
* :class:`ViewMenu`: View menu for the main menubar
* :class:`FormatMenu`: Format menu for the main menubar
* :class:`MacroMenu`: Macro menu for the main menubar
* :class:`HelpMenu`: Help menu for the main menubar
* :class:`FileHistoryMenu`: Menu showing recent files
* :class:`BorderChoiceMenu`: QMenu for choosing cell borders
* :class:`BorderWidthMenu`: QMenu for choosing the cell border width

"""

from functools import partial
from pathlib import Path

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
        QMenuBar, QMenu, QAction, QMainWindow, QWidget, QToolBar)

try:
    import matplotlib.figure as matplotlib_figure
except ImportError:
    matplotlib_figure = None

try:
    from pyspread.actions import MainWindowActions
    from pyspread.icons import Icon
except ImportError:
    from actions import MainWindowActions
    from icons import Icon


class MenuBar(QMenuBar):
    """The main menubar """

    def __init__(self, main_window: QMainWindow):
        """
        :param main_window: Application main window

        """

        super().__init__()

        self.main_window = main_window
        actions = main_window.main_window_actions

        self.file_menu = FileMenu(self, actions)
        self.edit_menu = EditMenu(self, actions)
        self.view_menu = ViewMenu(self, actions)
        self.format_menu = FormatMenu(self, actions)
        self.macro_menu = MacroMenu(self, actions)
        self.help_menu = HelpMenu(self, actions)

        self.addMenu(self.file_menu)
        self.addMenu(self.edit_menu)
        self.addMenu(self.view_menu)
        self.addMenu(self.format_menu)
        self.addMenu(self.macro_menu)
        self.addMenu(self.help_menu)


class FileMenu(QMenu):
    """File menu for the main menubar"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&File', parent)

        self.parent = parent

        self.addAction(actions.new)
        self.addAction(actions.open)
        self.history_submenu = FileHistoryMenu(self, actions)
        self.history_action = self.addMenu(self.history_submenu)
        self.addSeparator()
        self.addAction(actions.save)
        self.addAction(actions.save_as)
        self.addSeparator()
        self.addAction(actions.imprt)
        self.addAction(actions.export)
        self.addSeparator()
        self.addAction(actions.approve)
        self.addSeparator()
        self.addAction(actions.clear_globals)
        self.addSeparator()
        self.addAction(actions.print_preview)
        self.addAction(actions.print)
        self.addSeparator()
        self.addAction(actions.preferences)
        self.addSeparator()
        self.addAction(actions.quit)


class EditMenu(QMenu):
    """Edit menu for the main menubar"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&Edit', parent)

        self.addAction(actions.undo)
        self.addAction(actions.redo)
        self.addSeparator()
        self.addAction(actions.cut)
        self.addAction(actions.copy)
        self.addAction(actions.copy_results)
        self.addAction(actions.paste)
        self.addAction(actions.paste_as)
        self.addSeparator()
        self.addAction(actions.find)
        self.addAction(actions.replace)
        self.addSeparator()
        self.addAction(actions.sort_ascending)
        self.addAction(actions.sort_descending)
        self.addSeparator()
        self.addAction(actions.toggle_selection_mode)
        self.addSeparator()
        self.addAction(actions.quote)
        self.addSeparator()
        self.addAction(actions.insert_rows)
        self.addAction(actions.insert_columns)
        self.addAction(actions.insert_table)
        self.addSeparator()
        self.addAction(actions.delete_rows)
        self.addAction(actions.delete_columns)
        self.addAction(actions.delete_table)
        self.addSeparator()
        self.addAction(actions.resize_grid)


class ViewMenu(QMenu):
    """View menu for the main menubar"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&View', parent)

        self.addAction(actions.fullscreen)
        self.addSeparator()

        self.toolbar_submenu = self.addMenu('Toolbars')
        self.toolbar_submenu.addAction(actions.toggle_main_toolbar)
        self.toolbar_submenu.addAction(actions.toggle_macro_toolbar)
        self.toolbar_submenu.addAction(actions.toggle_format_toolbar)
        self.toolbar_submenu.addAction(actions.toggle_find_toolbar)

        self.addAction(actions.toggle_entry_line_dock)
        self.addAction(actions.toggle_macro_dock)
        self.addSeparator()
        self.addAction(actions.goto_cell)
        self.addSeparator()
        self.addAction(actions.toggle_spell_checker)
        self.addSeparator()
        self.addAction(actions.zoom_in)
        self.addAction(actions.zoom_out)
        self.addAction(actions.zoom_1)
        self.addSeparator()
        self.addAction(actions.refresh_cells)
        self.addAction(actions.toggle_periodic_updates)
        self.addSeparator()
        self.addAction(actions.show_frozen)


class FormatMenu(QMenu):
    """Format menu for the main menubar"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&Format', parent)

        self.addAction(actions.copy_format)
        self.addAction(actions.paste_format)
        self.addSeparator()
        self.addAction(actions.font)
        self.addAction(actions.bold)
        self.addAction(actions.italics)
        self.addAction(actions.underline)
        self.addAction(actions.strikethrough)
        self.addSeparator()

        self.renderer_submenu = self.addMenu('Cell renderer')
        self.renderer_submenu.addAction(actions.text)
        self.renderer_submenu.addAction(actions.image)
        self.renderer_submenu.addAction(actions.markup)
        if matplotlib_figure is not None:
            self.renderer_submenu.addAction(actions.matplotlib)

        self.addAction(actions.freeze_cell)
        self.addAction(actions.lock_cell)
        self.addAction(actions.button_cell)

        self.addSeparator()

        self.addAction(actions.merge_cells)

        self.addSeparator()

        self.rotation_submenu = self.addMenu('Rotation')
        self.rotation_submenu.addAction(actions.rotate_0)
        self.rotation_submenu.addAction(actions.rotate_90)
        self.rotation_submenu.addAction(actions.rotate_180)
        self.rotation_submenu.addAction(actions.rotate_270)

        self.justification_submenu = self.addMenu('Justification')
        self.justification_submenu.addAction(actions.justify_left)
        self.justification_submenu.addAction(actions.justify_center)
        self.justification_submenu.addAction(actions.justify_right)
        self.justification_submenu.addAction(actions.justify_fill)

        self.alignment_submenu = self.addMenu('Alignment')
        self.alignment_submenu.addAction(actions.align_top)
        self.alignment_submenu.addAction(actions.align_center)
        self.alignment_submenu.addAction(actions.align_bottom)

        self.addSeparator()

        self.border_submenu = BorderChoiceMenu(actions)
        self.addMenu(self.border_submenu)

        self.line_width_submenu = BorderWidthMenu(actions)
        self.addMenu(self.line_width_submenu)

        self.addSeparator()
        self.addAction(actions.text_color)
        self.addAction(actions.line_color)
        self.addAction(actions.background_color)


class MacroMenu(QMenu):
    """Macro menu for the main menubar"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&Macro', parent)

        self.addAction(actions.insert_image)
        if matplotlib_figure is not None:
            self.addAction(actions.insert_chart)
        self.addSeparator()
        self.addAction(actions.insert_sum)

class HelpMenu(QMenu):
    """Help menu for the main menubar"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&Help', parent)

        self.addAction(actions.manual)
        self.addAction(actions.tutorial)
        self.addSeparator()
        self.addAction(actions.dependencies)
        self.addSeparator()
        self.addAction(actions.about)


class FileHistoryMenu(QMenu):
    """Menu that displays the file history"""

    def __init__(self, parent: QWidget, actions: MainWindowActions):
        """
        :param parent: Parent widget
        :param actions: Main window actions

        """

        super().__init__('&Recent files', parent)

        self.main_window = parent.parent.main_window

    def update(self):
        """Updates file history menu"""

        self.clear()

        settings = self.main_window.settings
        for posixpath in settings.file_history[:settings.max_file_history]:
            filepath = Path(posixpath)
            if filepath.is_file():
                action = QAction(filepath.name, self)
                action.setStatusTip(posixpath)
                self.addAction(action)
                action.triggered.connect(self.on_recent)

    def on_recent(self):
        """Event handler for file history menu"""

        posixpath = self.sender().statusTip()
        self.main_window.workflows.file_open_recent(posixpath)


class BorderChoiceMenu(QMenu):
    """QMenu for choosing cell borders that shall be manipulated"""

    def __init__(self, actions: MainWindowActions):
        """
        :param actions: Main window actions

        """

        self.actions = actions

        super().__init__()

        self.setTitle("Formatted borders")
        self.setIcon(Icon.border_menu)

        self.addAction(actions.format_borders_all)
        self.addAction(actions.format_borders_top)
        self.addAction(actions.format_borders_bottom)
        self.addAction(actions.format_borders_left)
        self.addAction(actions.format_borders_right)
        self.addAction(actions.format_borders_outer)
        self.addAction(actions.format_borders_inner)
        self.addAction(actions.format_borders_top_bottom)

        self.triggered.connect(self.on_selection)

    def on_selection(self, event):
        """Event handler"""

        self.actions.parent.format_toolbar.border_menu_button.setAttribute(
            Qt.WA_UnderMouse, False)


class BorderWidthMenu(QMenu):
    """QMenu for choosing the cell border width"""

    def __init__(self, actions: MainWindowActions):
        """
        :param actions: Main window actions

        """

        self.actions = actions

        super().__init__()

        self.setTitle("Border width")
        self.setIcon(Icon.format_borders)

        self.addAction(actions.format_borders_0)
        self.addAction(actions.format_borders_1)
        self.addAction(actions.format_borders_2)
        self.addAction(actions.format_borders_4)
        self.addAction(actions.format_borders_8)
        self.addAction(actions.format_borders_16)
        self.addAction(actions.format_borders_32)
        self.addAction(actions.format_borders_64)

        self.triggered.connect(self.on_selection)

    def on_selection(self, event):
        """Event handler"""

        self.actions.parent.format_toolbar.border_menu_button.setAttribute(
            Qt.WA_UnderMouse, False)


class GridContextMenu(QMenu):
    """Context menu for grid"""

    def __init__(self, actions: MainWindowActions):
        """
        :param actions: Main window actions

        """

        super().__init__()

        self.addAction(actions.cut)
        self.addAction(actions.copy)
        self.addAction(actions.copy_results)
        self.addAction(actions.copy_format)
        self.addAction(actions.paste)
        self.addAction(actions.paste_as)
        self.addAction(actions.paste_format)
        self.addSeparator()
        self.addAction(actions.toggle_selection_mode)
        self.addSeparator()
        self.addAction(actions.copy_format)
        self.addAction(actions.paste_format)
        self.addSeparator()
        self.addAction(actions.toggle_selection_mode)
        self.addSeparator()
        self.addAction(actions.quote)
        self.addSeparator()
        self.addAction(actions.insert_rows)
        self.addAction(actions.insert_columns)
        self.addAction(actions.insert_table)
        self.addSeparator()
        self.addAction(actions.delete_rows)
        self.addAction(actions.delete_columns)
        self.addAction(actions.delete_table)


class HorizontalHeaderContextMenu(QMenu):
    """Context menu for horizontal grid header"""

    def __init__(self, actions: MainWindowActions):
        """
        :param actions: Main window actions

        """

        super().__init__()

        self.addAction(actions.insert_columns)
        self.addAction(actions.delete_columns)


class VerticalHeaderContextMenu(QMenu):
    """Context menu for vertical grid header"""

    def __init__(self, actions: MainWindowActions):
        """
        :param actions: Main window actions

        """

        super().__init__()

        self.addAction(actions.insert_rows)
        self.addAction(actions.delete_rows)


class TableChoiceContextMenu(QMenu):
    """Context menu for table choice"""

    def __init__(self, actions: MainWindowActions):
        """
        :param actions: Main window actions

        """

        super().__init__()

        self.addAction(actions.insert_table)
        self.addAction(actions.delete_table)


class ToolbarManagerMenu(QMenu):
    """Menu with all actions of a toolbar that allows toggling visibility"""

    def __init__(self, toolbar: QToolBar):
        """
        :param toolbar: Toolbar that the menu is managing

        """

        super().__init__()
        self.toolbar = toolbar

        for action in toolbar.actions():
            if action.isSeparator():
                self.addSeparator()
            else:
                self.addAction(self._get_toggle_action(action))
        self.update_checked_states()

    def _get_toggle_action(self, action: QAction) -> QAction:
        """Returns a toggle action for a toolbar action

        :param action: Toolbar action for which the toggle state is returned

        """

        taction = QAction(action.icon(), action.text(), action, checkable=True)
        taction.triggered.connect(partial(self.on_toggled, action))

        return taction

    def update_checked_states(self):
        """Updates checked states"""

        for tool_action, action in zip(self.toolbar.actions(), self.actions()):
            action.setChecked(tool_action.isVisible())

    def on_toggled(self, action: QAction, toggled: bool):
        """Action toggle event handler

        :param action: Toolbar action for which visibility is toggled
        :param toggled: If False then action is set invisible and vice versa

        """

        action.setVisible(toggled)



fig = Figure()
ax = fig.add_axes([.2,.2, .7, .7])

# Fixing random state for reproducibility
numpy.random.seed(19680801)


N = 50
x1 = numpy.random.rand(N)
y1 = numpy.random.rand(N)

x2 = 0.5 + numpy.random.rand(N) * 2
y2 = numpy.random.rand(N) ** 2

ax.scatter(x1, y1, c='r', alpha=0.5)
ax.scatter(x2, y2, c='b', alpha=0.5)

ax.set(title='Scatter chart', xlabel="X", ylabel="Y")

fig



fig = Figure()
ax = fig.add_axes([.2,.2, .7, .7])

N_points = 100000
n_bins = 20

# Fixing random state for reproducibility
numpy.random.seed(19680801)

# Generate a normal distribution
x = numpy.random.randn(N_points)

# We can set the number of bins with the `bins` kwarg
ax.hist(x, bins=n_bins)

ax.set(xlabel='Value', ylabel='Frequency',
       title='Histogram chart')

fig



from rpy2.interactive import process_revents
from rpy2.robjects.packages import importr
from rpy2.robjects.vectors import IntVector
from rpy2.robjects.lib import grdevices

try:
    process_revents.start()
except:
    pass

graphics = importr("graphics")

with grdevices.render_to_bytesio(grdevices.svg) as svg:
    graphics.barplot(IntVector((1, 6, 3, 5, 4, 1)), ylab="Value")
process_revents.stop()
svg.getvalue()



from mpl_toolkits.axes_grid1 import make_axes_locatable

fig = Figure()
ax = fig.add_axes([.2,.2, .7, .7])

# Fixing random state for reproducibility
numpy.random.seed(19680801)


# the random data
x = numpy.random.randn(1000)
y = numpy.random.randn(1000)


# the scatter plot:
ax.scatter(x, y, alpha=0.08)
ax.set_aspect(1.)

# create new axes on the right and on the top of the current axes
# The first argument of the new_vertical(new_horizontal) method is
# the height (width) of the axes to be created in inches.
divider = make_axes_locatable(ax)
axHistx = divider.append_axes("top", 0.6, pad=0.1, sharex=ax)
axHisty = divider.append_axes("right", 0.6, pad=0.1, sharey=ax)

# Make some labels invisible
axHistx.xaxis.set_tick_params(labelbottom=False)
axHisty.yaxis.set_tick_params(labelleft=False)

# Determine nice limits by hand:
binwidth = 0.25
xymax = max(numpy.max(numpy.abs(x)), numpy.max(numpy.abs(y)))
lim = (int(xymax/binwidth) + 1)*binwidth

bins = numpy.arange(-lim, lim + binwidth, binwidth)
axHistx.hist(x, bins=bins)
axHisty.hist(y, bins=bins, orientation='horizontal')

# the xaxis of axHistx and yaxis of axHisty are shared with axScatter,
# thus there is no need to manually adjust the xlim and ylim of these
# axis.

axHistx.set_yticks([0, 50, 100])

axHisty.set_xticks([0, 50, 100])

ax.set(xlabel="X", ylabel="Y")
ax.set_title('Scatter histogram chart', pad=70)

fig
